"""Magnet tools."""

from __future__ import annotations

from typing import Sequence

import numpy
from scipy.special import comb

from ..lattice.elements import Element

__all__ = [
    "feeddown_polynomba",
    "feeddown_from_nth_order",
    "feeddown_pol_from_element",
]


def feeddown_pol_from_element(
    ele: Element, verbose: bool = False, **kwargs: dict[str, any]
) -> dict[str, numpy.ndarray]:
    """
    Return the feed down polynoms due to a transverse offset on an element.

    Parameters:
        ele: a ring element.
        verbose: prints additional info. Default: False.
        kwargs:
            offset:(x,y) horizontal and vertical offset in meters.
                Default taken from T2 or T1. Otherwise, zero.

    Returns:
        Dictionary with PolynomB and PolynomA feeddown components.

    Note: Thin lens approximation, T2=-T1.
          Bending angles are ignored.
    """
    # check verbose flags only once
    verboseprint = print if verbose else lambda *a, **k: None
    # handle cases where polynom a and b are not defined
    polb = numpy.array([])
    pola = numpy.array([])
    if hasattr(ele, "PolynomB"):
        polb = ele.PolynomB
    else:
        verboseprint(f"Element {ele.FamName} has no PolynomB")
    if hasattr(ele, "PolynomA"):
        pola = ele.PolynomA
    else:
        verboseprint(f"Element {ele.FamName} has no PolynomA")

    # check if user has set offsets
    xoffset, yoffset = kwargs.get("offset", (0, 0))
    # check if element has T1 and T2. Use one.
    if "offset" not in kwargs:
        if hasattr(ele, "T2"):
            xoffset = ele.T2[0]
            yoffset = ele.T2[2]
        elif hasattr(ele, "T1"):
            xoffset = -ele.T1[0]
            yoffset = -ele.T1[2]
        else:
            verboseprint(f"Element {ele.FamName} has no T1 or T2.")
    verboseprint(f"Using offsets xoffset={xoffset}, yoffset={yoffset}.")
    # Return the polynoms
    return feeddown_polynomba(polb=polb, pola=pola, xoffset=xoffset, yoffset=yoffset)


def feeddown_polynomba(
    polb: Sequence[float],
    pola: Sequence[float],
    xoffset: float = 0,
    yoffset: float = 0,
    verbose: bool = False,
) -> dict[str, numpy.ndarray]:
    """
    Return the feeddown due to a transverse offset.

    Parameters:
        polb: PolynomB.
        pola: PolynomA.
        xoffset: Default zero. Horizontal offset in meters.
        yoffset: Default zero. Vertical offset in meters.
        verbose: prints additional info.

    Returns:
        Dictionary with PolynomB and PolynomA feeddown components.

    Note:
        If only one polynom is passed, it is assumed to be PolynomB.
    """
    # check verbose flags only once
    verboseprint = print if verbose else lambda *a, **k: None

    # verify polynoms length
    maxorda = len(pola)
    maxordb = len(polb)
    maxord = max(maxorda, maxordb)
    polasum = numpy.zeros(max(0, maxord - 1))
    polbsum = numpy.zeros(max(0, maxord - 1))
    if maxorda == 0 and maxordb == 0:
        verboseprint("Both polynoms are zero.")
    else:
        polbpad = numpy.pad(
            polb, (0, maxord - maxordb), "constant", constant_values=(0, 0)
        )
        polapad = numpy.pad(
            pola, (0, maxord - maxorda), "constant", constant_values=(0, 0)
        )

        verboseprint(f"polb={polb},pola={pola}")
        verboseprint(f"xoffset={xoffset},yoffset={yoffset}")
        for ith in range(2, maxord + 1):
            polbaux_b, polaaux_b = feeddown_from_nth_order(
                ith, polbpad[ith - 1], xoffset, yoffset, poltype="B"
            )
            polbaux_a, polaaux_a = feeddown_from_nth_order(
                ith, polapad[ith - 1], xoffset, yoffset, poltype="A"
            )
            polbshort = polbaux_b + polbaux_a
            polashort = polaaux_b + polaaux_a
            polbsum = polbsum + numpy.pad(
                polbshort, (0, maxord - ith), "constant", constant_values=(0, 0)
            )
            polasum = polasum + numpy.pad(
                polashort, (0, maxord - ith), "constant", constant_values=(0, 0)
            )
    poldict = {"PolynomB": polbsum, "PolynomA": polasum}
    verboseprint(f"poldict={poldict}")
    return poldict


def feeddown_from_nth_order(
    nthorder: int,
    nthpolcomp: float,
    xoffset: float,
    yoffset: float,
    poltype: str = "B",
    verbose: bool = False,
) -> tuple[numpy.ndarray, numpy.ndarray]:
    """
    Return the feeddown polynoms from an nth-order magnet component transverse offset.

    Parameters:
        nthorder: integer order of the magnet component, e.g. 1 for dipole,
            2 for quadrupoles, 3 for sextupoles, etc.
        nthpolcomp: float. nth component of the polynom, i.e. the value of
            PolynomA/B[nthorder-1].
        xoffset: float. Horizontal offset in meters.
        yoffset: float. Vertical offset in meters.
        poltype: Default 'B'. Could be 'B' or 'A', otherwise ignored.
        verbose: print info on input and output parameters.

    Returns:
        Tuple of two numpy arrays with PolynomB and PolynomA from feeddown.
    """
    verboseprint = print if verbose else lambda *a, **k: None
    verboseprint(f"nthorder={nthorder},nthpolcomp={nthpolcomp},poltype={poltype}")
    verboseprint(f"xoffset={xoffset},yoffset={yoffset}")
    fakeimag = {0: 1, 1: 0, 2: -1, 3: 0}
    polbaux = numpy.zeros(nthorder - 1)
    polaaux = numpy.zeros(nthorder - 1)
    if nthpolcomp != 0:
        for kterm in range(1, nthorder):
            ichoosek = comb(nthorder - 1, kterm)
            pascalsn = comb(kterm, numpy.arange(kterm + 1))
            powk = numpy.arange(kterm + 1)
            powkflip = numpy.arange(kterm, -1, -1)
            recoefs = numpy.array([fakeimag[numpy.mod(idx, 4)] for idx in powk])
            imcoefs = numpy.array([fakeimag[numpy.mod(idx + 3, 4)] for idx in powk])
            commonfactor = nthpolcomp * ichoosek * (-1) ** kterm
            repart = (
                recoefs
                * commonfactor
                * (pascalsn * (xoffset**powkflip * yoffset**powk))
            )
            impart = (
                imcoefs
                * commonfactor
                * (pascalsn * (xoffset**powkflip * yoffset**powk))
            )
            polbaux[nthorder - kterm - 1] = polbaux[nthorder - kterm - 1] + repart.sum()
            polaaux[nthorder - kterm - 1] = polaaux[nthorder - kterm - 1] + impart.sum()
    verboseprint(f"polbaux={polbaux},polaaux={polaaux}")
    polbout, polaout = polbaux, polaaux
    # skew components swap the imaginary and real feed-down and change the sign
    # of the real part, i.e. j*j = -1
    if poltype == "A":
        polbout, polaout = -1 * polaaux, polbaux
    return polbout, polaout



"""Matching of lattice parameters.

The matching acts on a :py:class:`.VariableList` to statisfy the constraints
expressed in an :py:class:`.ObservableList`.

Examples of use of such classes are available in:

* :doc:`/p/notebooks/variables`
* :doc:`/p/notebooks/observables`

Examples of matching are given in :doc:`/p/notebooks/matching`.
"""

from __future__ import annotations

__all__ = ["match"]


import numpy as np
from scipy.optimize import least_squares

from .observablelist import ObservableList
from ..lattice import Lattice, VariableList
from ..lattice.lattice_variables import ElementVariable


def _match(
    ring: Lattice,
    variables: VariableList,
    constraints: ObservableList,
    *,
    method: str | None = None,
    verbose: int = 2,
    max_nfev: int = 1000,
    dp: float | None = None,
    dct: float | None = None,
    df: float | None = None,
    **kwargs,
) -> None:
    def fun(vals):
        """Evaluation function for the minimiser."""
        variables.set(vals, ring=ring)
        constraints.evaluate(ring, dp=dp, dct=dct, df=df)
        return constraints.get_flat_weighted_deviations(err=1.0e6)

    vini = variables.get(ring=ring, initial=True, check_bounds=True)
    bounds = np.array([var.bounds for var in variables])

    constraints.evaluate(ring, initial=True, dp=dp, dct=dct, df=df)
    constraints.check()
    ntargets = constraints.flat_values.size

    if method is None:
        if np.any(np.isfinite(bounds)) or ntargets < len(variables):
            method = "trf"
        else:
            method = "lm"
    if verbose >= 1:
        print(
            f"\n{ntargets} constraints, {len(variables)} variables,"
            f" using method {method}\n"
        )

    least_squares(
        fun,
        vini,
        bounds=(bounds[:, 0], bounds[:, 1]),
        verbose=verbose,
        max_nfev=max_nfev,
        method=method,
        **kwargs,
    )

    if verbose >= 1:
        print("\nConstraints:")
        print(constraints)
        print("\nVariables:")
        print(variables)


def match(
    ring: Lattice,
    variables: VariableList,
    constraints: ObservableList,
    *,
    copy: bool = False,
    **kwargs,
) -> Lattice | None:
    """Match constraints by varying variables.

    Minimisation is performed by the :py:func:`~scipy.optimize.least_squares`
    function.

    Args:
        ring:           Lattice description
        variables:      Variable parameters
        constraints:    Constraints to fulfill
        copy:           If :py:obj:`True`, return a modified copy of *ring*, otherwise
          perform the match in-line

    Keyword Args:
        method:         Minimisation algorithm (see
          :py:func:`~scipy.optimize.least_squares`). If :py:obj:`None`, use
          'lm' for unbounded problems, 'trf' otherwise.
        verbose:        Level of verbosity
        max_nfev:       Maximum number of function evaluation
        dp:             Momentum deviation.
        dct:            Path lengthening.
        df:             Deviation from the nominal RF frequency.
        **kwargs:       Keyword arguments sent to
          :py:func:`~scipy.optimize.least_squares`

    Returns:
        Modified Lattice if copy=True, None otherwise

    Raises:
        TypeError: If copy=True and ElementVariable is present
        RuntimeError: If optimisation fails

    Examples:

        See :doc:`/p/notebooks/matching`
    """
    if copy:
        # Check that there is no ElementVariable
        for var in variables:
            if isinstance(var, ElementVariable):
                raise TypeError("When 'copy' is True, no ElementVariable is accepted")
        newring = ring.deepcopy()
        _match(newring, variables, constraints, **kwargs)
        return newring
    else:
        _match(ring, variables, constraints, **kwargs)
        return None



r"""Using `MAD8`_ files with PyAT
=================================

Using MAD8 files is similar to using MAD-X files: see
:py:mod:`the MAD-X documentation <at.load.madx>`

.. _mad8: https://mad8.web.cern.ch/user/mad.html
"""

from __future__ import annotations

__all__ = ["Mad8Parser", "load_mad8", "save_mad8"]

# functions known by MAD-8
from math import pi, e, sqrt, exp, log, sin, cos, tan  # noqa: F401
from math import asin  # noqa: F401
import re

# constants known by MAD-8
from scipy.constants import c as clight, e as qelect  # noqa: F401
from scipy.constants import physical_constants as _cst

from ..lattice import Lattice

from .file_input import ignore_names

# noinspection PyProtectedMember
from .madx import _MadElement, _MadParser, _MadExporter

# Commands known by MAD8
# noinspection PyProtectedMember
from .madx import (  # noqa: F401
    drift,
    marker,
    quadrupole,
    sextupole,
    octupole,
    multipole,
    sbend,
    rbend,
    kicker,
    hkicker,
    vkicker,
    rfcavity,
    monitor,
    hmonitor,
    vmonitor,
    value,
)

# Constants known by MAD-8
_true_ = True
_yes_ = True
_t_ = True
_on_ = True
_false_ = False
_no_ = False
_f_ = False
_off_ = False
twopi = 2 * pi
degrad = 180.0 / pi
raddeg = pi / 180.0
emass = 1.0e-03 * _cst["electron mass energy equivalent in MeV"][0]  # [GeV]
pmass = 1.0e-03 * _cst["proton mass energy equivalent in MeV"][0]  # [GeV]

_attr = re.compile(r"\[([a-zA-Z_][\w.:]*)]")  # Identifier enclosed in square brackets

ignore_names(
    globals(),
    _MadElement,
    ["solenoid", "rfmultipole", "crabcavity", "elseparator", "collimator", "tkicker"],
)


class Mad8Parser(_MadParser):
    # noinspection PyUnresolvedReferences
    r"""MAD-X specific parser

    The parser is a subclass of :py:class:`dict` and is database containing all the
    MAD-X variables.

    Example:
        Parse a 1\ :sup:`st` file:

        >>> parser = at.Mad8Parser()
        >>> parser.parse_file("file1")

        Parse another file:

        >>> parser.parse_file("file2")

        Get the variable "vkick"

        >>> parser["vkick"]
        0.003

        Define a new variable:

        >>> parser["hkick"] = -0.0024

        Get the "qf1" element

        >>> parser["qf1"]
        quadrupole(name=qf1, l=1.0, k1=0.5, tilt=0.001)

        Generate an AT :py:class:`.Lattice` from the "ring" sequence

        >>> ring = parser.lattice(use="ring")  # generate an AT Lattice
    """

    _continuation = "&"
    _blockcomment = ("comment", "endcomment")

    def __init__(self, **kwargs):
        """
        Args:
            strict:     If :py:obj:`False`, assign 0 to undefined variables
            verbose:    If :py:obj:`True`, print details on the processing
            **kwargs:   Initial variable definitions
        """
        super().__init__(globals(), **kwargs)

    def _format_command(self, expr: str) -> str:
        """Evaluate an expression using *self* as local namespace"""
        expr = _attr.sub(r".\1", expr)  # Attribute access: VAR[ATTR]
        expr = expr.replace("^", "**")  # Exponentiation
        return super()._format_command(expr)


def load_mad8(
    *files: str,
    use: str = "ring",
    strict: bool = True,
    verbose: bool = False,
    **kwargs,
) -> Lattice:
    """Create a :py:class:`.Lattice` from MAD8 files

    - The *energy* and *particle* of the generated lattice are taken from the MAD8
      ``BEAM`` object, using the MAD8 default parameters: positrons at 1 Gev.
      These parameters are overloaded by the value given in the *energy* and
      *particle* keyword arguments.
    - The radiation state is given by the ``RADIATE`` flag of the ``BEAM`` object,
      using the AT defaults: RF cavities active, synchrotron radiation in dipoles and
      quadrupoles.
    - Long elements are split according to the default AT value for *NumIntSteps* (10).

    Parameters:
        files:              Names of one or several MAD8 files
        strict:             If :py:obj:`False`, assign 0 to undefined variables
        use:                Name of the MAD8 sequence or line containing the desired
          lattice. Default: ``ring``
        verbose:            If :py:obj:`True`, print details on the processing

    Keyword Args:
        name (str):         Name of the lattice. Default: MAD8 sequence name.
        particle(Particle): Circulating particle. Default: from MAD8
        energy (float):     Energy of the lattice [eV]. Default: from MAD8
        periodicity(int):   Number of periods. Default: 1
        *:                  Other keywords will be used as initial variable definitions

    Returns:
        lattice (Lattice):  New :py:class:`.Lattice` object
    """
    parser = Mad8Parser(strict=strict, verbose=verbose)
    params = {
        key: kwargs.pop(key)
        for key in ("name", "particle", "energy", "periodicity")
        if key in kwargs
    }
    parser.parse_files(*files, **kwargs)
    return parser.lattice(use=use, **params)


class _Mad8Exporter(_MadExporter):
    delimiter = ""
    continuation = "&"
    bool_fmt = {False: ".FALSE.", True: ".TRUE."}


def save_mad8(ring: Lattice, filename: str | None = None, **kwargs):
    """Save a :py:class:`.Lattice` as a MAD8 file

    Args:
        ring:   lattice
        filename: file to be created. If None, write to sys.stdout

    Keyword Args:
        use (str | None): name of the created SEQUENCE of LINE.
          Default: name of the PyAT lattice
        use_line (bool):  If True, use a MAD "LINE" format. Otherwise, use
          a MAD "SEQUENCE"
    """
    exporter = _Mad8Exporter(ring, **kwargs)
    exporter.export(filename)



from __future__ import annotations

__all__ = ["RDTObservable"]

from functools import partial

import numpy as np

from .observables import Need, ElementObservable
from ..lattice import Refpts
from ..physics import RDTType

RDT_names = {
    RDTType.FOCUSING: ("h20000", "h00200"),
    RDTType.COUPLING: ("h10010", "h10100"),
    RDTType.CHROMATIC: ("h11001", "h00111", "h20001", "h00201", "h10002"),
    RDTType.GEOMETRIC1: ("h21000", "h30000", "h10110", "h10020", "h10200"),
    RDTType.GEOMETRIC2: (
        "h22000",
        "h11110",
        "h00220",
        "h31000",
        "h40000",
        "h20110",
        "h11200",
        "h20020",
        "h20200",
        "h00310",
        "h00400",
    ),
    RDTType.TUNESHIFT: ("dnux_dJx", "dnux_dJy", "dnuy_dJy"),
}
RDT_code = {nm: code for code, names in RDT_names.items() for nm in names}


def _rdt_access(param, data):
    return getattr(data, param)


class RDTObservable(ElementObservable):
    """Observe a resonance driving term at selected locations.

    Process the local output of :py:func:`.get_rdts`.
    """

    _rdt_type: RDTType

    def __init__(
        self,
        refpts: Refpts,
        param: str,
        name: str | None = None,
        second_order: bool = False,
        **kwargs,
    ):
        # noinspection PyUnresolvedReferences
        r"""Args:
            refpts:         Observation points.
              See ":ref:`Selecting elements in a lattice <refpts>`"
            param:          :ref:`RDT name <rdt_param>`
            name:           Observable name. If :py:obj:`None`, an explicit
              name will be generated
            second_order:   Compute second order terms. Computation is significantly
              longer using this method

        Keyword Args:
            procfun:        Post-processing function. It can be any numpy ufunc or a
              function name in {"real", "imag", "abs", "angle", "log", "exp", "sqrt"}.
            statfun:        Statistics post-processing function. it can be a numpy
              function or a function name in {"mean", "std", "var", "min", "max"}.
              Example: :pycode:`statfun=numpy.mean`.
            target:         Target value for a constraint. If :py:obj:`None`
              (default), the residual will always be zero.
            weight:         Weight factor: the residual is
              :pycode:`((value-target)/weight)**2`
            bounds:         Tuple of lower and upper bounds. The parameter
              is constrained in the interval
              [*target*\ +\ *low_bound* *target*\ +\ *up_bound*]

        The *target*, *weight* and *bounds* inputs must be broadcastable to the
        shape of *value*.

        .. _rdt_param:
        .. rubric:: RDT parameter names

        **h20000**          at.RDTType.FOCUSING
        **h00200**

        **h10010**          at.RDTType.COUPLING
        **h10100**

        **h11001**          at.RDTType.CHROMATIC
        **h00111**
        **h20001**
        **h00201**
        **h10002**

        **h21000**          at.RDTType.GEOMETRIC1
        **h30000**
        **h10110**
        **h10020**
        **h10200**

        **h22000**          at.RDTType.GEOMETRIC2
        **h11110**
        **h00220**
        **h31000**
        **h40000**
        **h20110**
        **h11200**
        **h20020**
        **h20200**
        **h00310**
        **h00400**

        **dnux_dJx**        at.RDTType.DETUNING
        **dnux_dJy**
        **dnuy_dJy**

        Examples:

            >>> obs = RDTObservable(at.Monitor, "h20000")

            Observe "h20000" at all :py:class:`.Monitor` locations
        """

        needs = {Need.RDT}
        if second_order:
            needs.add(Need.RDT_2ND_ORDER)
        name = self._set_name(name, param, None)
        fun = partial(_rdt_access, param)

        super().__init__(fun, refpts, needs=needs, name=name, **kwargs)
        self._rdt_type = RDT_code[param]



"""Generic text parsers for conversion of lattices in different formats to AT"""

from __future__ import annotations

__all__ = [
    "set_argparser",
    "skip_class",
    "ignore_class",
    "skip_names",
    "ignore_names",
    "AnyDescr",
    "ElementDescr",
    "SequenceDescr",
    "BaseParser",
    "UnorderedParser",
    "LowerCaseParser",
    "UpperCaseParser",
    "DictNoDot",
]

from os import getcwd
from os.path import join, abspath, normpath, dirname
import re
from itertools import repeat, count
from collections.abc import Callable, Iterable, Generator, Mapping, Sequence

import numpy as np

from .utils import split_ignoring_parentheses, protect, restore
from ..lattice import Lattice, elements as elt, params_filter

_dot = re.compile(r'("?)(\.?[a-zA-Z_][\w.:]*)\1')  # look for MAD identifiers
_colon = re.compile(r":(?!=)")  # split on :=

# For decoding error messages:
_singlequoted = re.compile(r"'([\w.]*)'")  # look for single-quoted items
_named = re.compile(r"name=([\w.]*)")  # look for 'name=MADid' items


def set_argparser(argparser):
    """Decorator which adds an "argparser" attribute to a function"""

    def decorator(func):
        func.argparser = argparser
        return func

    return decorator


def skip_class(classname: str, baseclass: type[ElementDescr], **kwargs):
    """Generate a class for skipped elements.

    Args:
        classname: Name of the generated class
        baseclass: Base class, must be a subclass of :py:class:`ElementDescr`
        **kwargs: dictionary of additional attributes and methods. See :py:func:`type`.

    Returns:
        cls: Element class, skipped when generating AT elements
    """

    def init(self, *args, **kwargs):
        baseclass.__init__(self, *args, **kwargs)
        # if type(self) not in self.mentioned:
        type1 = self.__class__.__name__
        print(f"Element {self.name} ({type1}) is ignored.")
        self._mentioned.add(type(self))

    kwargs.update(__init__=init)
    return type(classname, (baseclass,), kwargs)


def ignore_class(classname: str, baseclass: type[ElementDescr], **kwargs):
    """Generate a class for ignored elements.

    Args:
        classname: Name of the generated class
        baseclass: Base class, must be a subclass of :py:class:`ElementDescr`
        **kwargs: dictionary of additional attributes and methods. See :py:func:`type`.

    Returns:
        cls: Element class, converted to :py:class:`.Marker` or :py:class:`.Drift`
          when generating AT elements
    """

    def init(self, *args, **kwargs):
        baseclass.__init__(self, *args, **kwargs)
        # if type(self) not in self.mentioned:
        if not args:  # not for replication
            type1 = self.__class__.__name__
            type2 = "Marker" if self.get("l", 0.0) == 0.0 else "Drift"
            print(f"Element {self.name} ({type1}) is replaced by a {type2}.")
            self._mentioned.add(type(self))

    def to_at(self, l=0.0, origin="", **params):  # noqa: E741
        if l == 0.0:
            return [elt.Marker(self.name, origin=origin, **self.meval(params))]
        else:
            return [elt.Drift(self.name, l, origin=origin, **self.meval(params))]

    kwargs.update(__init__=init, to_at=to_at)
    return type(classname, (baseclass,), kwargs)


def ignore_names(
    namespace: dict, baseclass: type[ElementDescr], classnames: Iterable[str]
) -> None:
    """Add classes for ignored elements in the given namespace.

    Args:
        namespace:
        baseclass:  Base class, must be a subclass of :py:class:`ElementDescr`
        classnames: Class names of the generated classes
    """
    namespace.update((nm, ignore_class(nm, baseclass)) for nm in classnames)


def skip_names(
    namespace: dict, baseclass: type[ElementDescr], classnames: Iterable[str]
) -> None:
    """Add classes for ignored elements in the given namespace.

    Args:
        namespace:
        baseclass:  Base class, must be a subclass of :py:class:`ElementDescr`
        classnames: Class names of the generated classes
    """
    namespace.update((nm, skip_class(nm, baseclass)) for nm in classnames)


class DictNoDot(dict):
    @classmethod
    def _defkey(cls, expr: str, quoted: bool) -> str:
        """substitutions to get a valid python identifier"""
        # Using classmethod to allow using super() in subclasses
        return expr.replace(".", "_").replace(":", "_")

    @classmethod
    def _gen_key(cls, expr: str) -> str:
        """Generate a dict key"""
        if expr and expr[0] == '"':
            return cls._defkey(expr[1:-1], True)
        else:
            return cls._defkey(expr, False)

    @classmethod
    def _gen_expr(cls, expr) -> str:
        """Generate a valid python expression"""

        def repl(match):
            return cls._defkey(match.group(2), match.group(1))

        return _dot.sub(repl, expr)

    def __init__(self, *args, verbose: bool = False, **kwargs):
        super().__init__(*args, **kwargs)
        self.verbose = verbose

    def __setitem__(self, key, value):
        super().__setitem__(self._gen_key(key), value)

    def __getitem__(self, key):
        return super().__getitem__(self._gen_key(key))

    def get(self, key, *args):
        return super().get(self._gen_key(key), *args)

    def _print(self, *args, **kwargs):
        if self.verbose:
            print(*args, **kwargs)


class AnyDescr:
    """Base class for source object descriptors"""

    str_attr = ()
    "list of names of str attributes"
    bool_attr = ()
    "list of names of bool attributes"
    pos_args = ()
    "list of names of positional arguments"

    @classmethod
    def argparser(cls, parser, argcount, argstr):
        """Specialised argument parser"""
        return parser._argparser(
            argcount, argstr, bool_attr=cls.bool_attr, str_attr=cls.str_attr
        )

    def __init__(self, *args, **kwargs):
        self.name = kwargs.pop("name", self.__class__.__name__)
        self.inverse = kwargs.pop("inverse", False)
        super().__init__(*args, **kwargs)

    def __neg__(self):
        return self.inverted(copy=True)

    def __call__(self, *args, copy: bool = True, **kwargs) -> AnyDescr | None:
        """Create a copy of the element with updated fields"""
        if copy:
            b = {key: kwargs.pop(key, value) for key, value in vars(self).items()}
            b.update(kwargs)
            return type(self)(self, *args, **b)
        else:
            self.update(*args, **kwargs)
            return None

    def update(self, *args, **kwargs):
        # Update attributes
        for key, value in vars(self).items():
            setattr(self, key, kwargs.pop(key, value))
        if isinstance(self, Mapping):
            # Update mapping
            # noinspection PyUnresolvedReferences
            super().update(*args, **kwargs)
        else:
            # Add new attributes
            for key, value in kwargs:
                setattr(self, key, value)

    def inverted(self, copy=False):
        """Return a reversed element or line"""
        instance = self() if copy else self
        instance.inverse = not self.inverse
        return instance

    def expand(self, parser: BaseParser) -> Generator[elt.Element, None, None]:
        """Iterator on the generated AT elements"""
        yield from ()


class ElementDescr(AnyDescr, dict):
    """Simple representation of an element as a :py:class:`dict`"""

    _mentioned = set()
    at2mad = {"Length": "L"}
    array_fmt = str.maketrans("[]()", "{}{}")
    bool_fmt = None

    def __init__(self, *args, origin=None, **kwargs):
        super().__init__(*args, **kwargs)
        self.origin = origin or self.__class__.__name__

    def __getattr__(self, item):
        # Allows accessing items using the attribute access syntax
        # tried after lookup for a real attribute failed
        try:
            return self[item]
        except KeyError as exc:
            # necessary for getattr to return the default value for a missing attribute
            name = self.__class__.__name__
            raise AttributeError(f"{name!r} object has no {item!r} attribute") from exc

    def __rmul__(self, other):
        """Element repetition"""
        return list(repeat(self, other))

    def __repr__(self):
        keywords = [f"name={self.name}"]
        keywords += [f"{k}={v!r}" for k, v in self.items()]
        return f"{self.__class__.__name__}({', '.join(keywords)})"

    def __str__(self):
        attrs = [f"{key}={self.attr_format(value)}" for key, value in self.items()]
        return ", ".join([self.__class__.__name__.upper().ljust(10)] + attrs)

    @staticmethod
    def attr_format(value):
        if isinstance(value, bool):
            return ElementDescr.bool_fmt[value]
        elif isinstance(value, np.ndarray):
            return np.array2string(value, separator=", ").translate(
                ElementDescr.array_fmt
            )
        elif isinstance(value, Sequence):
            return str(value).translate(ElementDescr.array_fmt)
        else:
            return str(value)

    @classmethod
    def from_at(cls, kwargs):
        def translate(attributes):
            for at, mad in cls.at2mad.items():
                v = attributes.pop(at, None)
                if v is not None:
                    yield mad, v

        params = {"name": kwargs.pop("FamName", "?")}
        params.update(translate(kwargs))
        return cls(**params)

    def to_at(self, *args, **params) -> list[elt.Element]:
        """Generate the AT element. Must be overloaded for each specific element"""
        return []

    # noinspection PyUnusedLocal
    def expand(self, parser: BaseParser) -> Generator[elt.Element, None, None]:
        """Iterator on the generated AT elements"""
        try:
            elems = self.to_at(**self)
        except Exception as exc:
            exc.args += (f"{self}",)
            raise

        if self.inverse:
            for elem in reversed(elems):
                yield elem.swap_faces(copy=True)
        else:
            yield from elems

    @property
    def length(self) -> float:
        """Element length"""
        return self.get("l", 0.0)

    @staticmethod
    def meval(params: dict):
        """Evaluation of superfluous parameters"""
        # return params
        return {}


class SequenceDescr(AnyDescr, list):
    """Simple representation of a sequence of elements as a :py:class:`list`"""

    def __repr__(self):
        string = super().__repr__()
        return f"{self.__class__.__name__}({string})"

    @property
    def length(self) -> float:
        return getattr(self, "l", 0.0)


class BaseParser(DictNoDot):
    """Generic file parser

    Analyses files with the following MAD-like format:

    ``variable = value``

    ``label : command [,attribute=value] [,attribute=value]...``

    The parser builds a database of all the defined objects
    """

    _delimiter: str | None = None
    _continuation: str = "\\"
    _linecomment: str | tuple[str] | None = "#"
    _blockcomment: tuple[str, str] | None = None
    _endfile: str | None = None
    _undef_key: str = "missing"

    def __init__(
        self,
        env: dict,
        *args,
        strict: bool = True,
        always_force: bool = True,
        **kwargs,
    ):
        """
        Args:
            env: global namespace used for evaluating commands
            verbose: If True, print detail on the processing
            strict: If :py:obj:`False`, assign 0 to undefined variables
            *args: dict initializer
            **kwargs: dict initializer
        """
        linecomment = self._linecomment
        blockcomment = self._blockcomment
        if isinstance(linecomment, tuple):

            def line_comment(line):
                for linecom in linecomment:
                    line, *_ = line.split(sep=linecom, maxsplit=1)
                return line

        else:
            if linecomment is None:

                def line_comment(line):
                    return line

            else:

                def line_comment(line):
                    line, *_ = line.split(sep=linecomment, maxsplit=1)
                    return line

        if blockcomment is None:
            # noinspection PyUnusedLocal
            def handle_comments(buffer, line, in_comment):
                line = line_comment(line)
                if line:
                    buffer.append(line_comment(line))
                    return False, ""
                else:
                    # Special case to avoid that empty lines break the continuation
                    return False, None

        else:

            def handle_comments(buffer, line, in_comment):
                if in_comment:
                    *rest, line = line.split(sep=endcomment, maxsplit=1)
                    in_comment = len(rest) <= 0
                    return in_comment, "" if in_comment > 0 else line
                else:
                    line = line_comment(line)
                    if line:
                        contents, *rest = line.split(sep=begcomment, maxsplit=1)
                        buffer.append(contents)
                        in_comment = len(rest) > 0
                        return in_comment, rest[0] if in_comment else ""
                    else:
                        # Special case to avoid that empty lines break the continuation
                        return False, None

            begcomment, endcomment = blockcomment

        self.skip_comments = handle_comments
        self.env = env
        self.bases = [getcwd()]
        self.kwargs = kwargs
        self.strict = strict
        self.always_force = always_force
        self.force = always_force

        super().__init__(*args, **kwargs)

        if not strict:
            self[self._undef_key] = 0
        self.postponed = []
        self.in_file = []

    def clear(self):
        """Clear the database: remove all parameters and objects"""
        super().clear()
        self.update(self.kwargs)
        if not self.strict:
            self[self._undef_key] = 0
        self.postponed = []
        self.in_file = []

    def _format_command(self, expr: str) -> str:
        """Format a command for evaluation

        Overload this method for specific languages"""
        return expr

    def evaluate(self, expr: str):
        """Evaluate the right side of an expression

        Args:
            expr: expression to evaluate

        Returns:
            value: evaluated expression
        """
        expr = self._format_command(self._gen_expr(expr))
        default_value = self.get(self._undef_key)
        if self.force and default_value is not None:
            for _loop in range(5):
                try:
                    return eval(expr, self.env, self)
                except NameError as exc:
                    var = self._reason(exc)
                    self._print(f"Set {var!r} to {default_value} ({_loop})")
                    self[var] = default_value
        return eval(expr, self.env, self)

    def _eval_cmd(self, cmdname: str, no_global: bool = False) -> Callable:
        """Evaluate a command"""
        cmd: Callable | None = self.get(cmdname, None)
        if cmd is not None:
            return cmd
        elif no_global:
            raise TypeError(f"{cmdname!r} is not allowed in this context")
        else:
            return self.env[self._gen_key(cmdname)]

    @staticmethod
    def _reason(exc: Exception) -> str | None:
        """Extract the element name from the exception"""
        if isinstance(exc, KeyError):  # Undefined element, attribute
            return exc.args[0]
        elif isinstance(exc, NameError):  # refpos missing
            return _singlequoted.search(exc.args[0])[1]
        elif isinstance(exc, TypeError):
            idx = _named.search(exc.args[-1])  # Missing pos. arg.
            if idx is None:
                idx = _singlequoted.search(exc.args[0])  # Not allowed in seq.
            if idx is None:
                return "TypeError"
            else:
                return idx[1]
        elif isinstance(exc, ValueError):  # overlap
            print(exc.args[0])
            names = _singlequoted.search(exc.args[0])
            return names[-1]
        else:
            return None

    def _argparser(
        self,
        argcount: int,
        argstr: str,
        *,
        bool_attr: tuple[str] = (),
        str_attr: tuple[str] = (),
        pos_args: tuple[str] = (),
    ):
        """Evaluate the value of a command argument and return the pair (key, value)"""

        def arg_value(k, v):
            if k in str_attr:
                return v[1:-1] if v[0] == '"' else v
            else:
                return self.evaluate(v)

        key, *value = split_ignoring_parentheses(
            argstr, delimiter="=", fence=('"', '"'), maxsplit=1
        )
        if value:  # Keyword argument
            return key, arg_value(key, value[0])
        else:
            ok = argstr[0] != "-"
            key = argstr if ok else argstr[1:]
            if key in bool_attr:  # boolean flag
                return key, ok
            else:  # positional parameter
                try:
                    key = pos_args[argcount]
                except IndexError:
                    print(f"Unexpected positional argument '{argstr}' ignored")
                    return None
                return key, arg_value(key, argstr)

    def _assign(self, label: str, key: str, value: str):
        """Variable assignment"""
        return key, self.evaluate(value)

    def _raw_command(
        self,
        label: str | None,
        cmdname: str,
        *args: str,
        no_global: bool = False,
        **kwargs,
    ):
        """Command execution"""
        # Get the command
        cmd = self._eval_cmd(cmdname, no_global=no_global)
        # Evaluate the arguments
        argp = getattr(cmd, "argparser", None)
        if argp:
            ags = (argp(self, n, arg) for n, arg in enumerate(args) if arg)
        else:
            ags = (self._argparser(n, arg) for n, arg in enumerate(args) if arg)
        kwargs.update(arg for arg in ags if arg is not None)
        if label is None:
            kwargs.setdefault("copy", False)
        else:
            kwargs.setdefault("name", label.replace('"', ""))
        # Execute the command
        return cmd(**kwargs)

    def _command(self, *args: str, **kwargs):
        return self._raw_command(*args, **kwargs)

    def _decode(self, label: str | None, cmdname: str, *args: str) -> None:
        """Execute the split statement"""
        left, *right = cmdname.split("=")
        try:
            if right:
                label, result = self._assign(label, left, right[0])
            else:
                result = self._command(label, cmdname, *args)
            if not (label is None or result is None):
                self[label] = result
        except (KeyError, NameError) as exc:  # store the failing assignment
            self._fallback(exc, label, cmdname, *args)

    def _fallback(self, exc: Exception, lbl: str | None, cmd: str, *args: str) -> None:
        """Store failing commands in self.postponed for later evaluation"""
        self.postponed.append((self._reason(exc), lbl, cmd, *args))

    def _format_statement(self, line: str) -> str:
        """Reformat the input line"""
        return line.replace(" ", "")  # Remove all spaces

    def _statement(self, line: str) -> bool:
        # protect quoted items. Make sure placeholder cannot be modified
        line, match1 = protect(line, fence=('"', '"'), placeholder="_0_")

        if self._endfile is not None and line.startswith(self._endfile):
            return False

        *left, right = _colon.split(line, maxsplit=1)
        right = self._format_statement(right)

        # protect nested parentheses
        cpt = count()
        rpt = []
        while "(" in right:
            plh = str(next(cpt)).join(("_par", "_"))
            right, match = protect(right, fence=("\\(", "\\)"), placeholder=plh)
            rpt.append(match)

        cmdargs = right.split(",")

        # restore parentheses
        for match in reversed(rpt):
            cmdargs = restore(match, *cmdargs)

        # restore quoted items
        b = restore(match1, *left, *cmdargs)

        # handle label
        if left:
            id = 1
            label = b[0].replace(" ", "")
        else:
            id = 0
            label = None

        # ignore MAD qualifiers
        while b[id] in ["const", "int", "real"]:
            id += 1

        # process statement
        self._decode(label, *b[id:])
        return True

    def _finalise(self, final: bool = True) -> None:
        """Called at the end of processing"""
        if final:
            undefined = self._missing(verbose=self.verbose)
            self._print(f"{len(undefined)} missing definitions.")

    @property
    def sequences(self):
        """List of available sequences or lines"""
        return [k for k, v in self.items() if isinstance(v, SequenceDescr)]

    @staticmethod
    def _command_str(label: str | None, cmdname: str, *argnames: str):
        string = ", ".join((cmdname, *argnames))
        if label is not None:
            string = " : ".join((label, string))
        return string

    def _lookup(self, item: str):
        """Search for an object in the pending statements"""
        for reason, label, *args in self.postponed:
            if label is not None and self._gen_key(label) == item:
                return (reason, label, *args)
        return None

    def _missing(self, verbose=False):
        """Return the set of missing definitions"""
        miss = set()
        for cmd in self.postponed:
            reason = cmd[0]
            if reason == self._gen_key(cmd[2]):
                if verbose:
                    cmdstr = self._command_str(*cmd[1:])
                    self._print(f"Command {cmdstr!r} ignored")
                continue
            while cmd is not None:
                reason = cmd[0]
                cmd = self._lookup(reason)
            miss.add(reason)
        return miss

    missing = property(_missing, doc="Set of missing definitions")

    def _analyse(self, key: str) -> None:
        """Print the chain of failing commands"""
        if isinstance(key, str):
            reason = self._gen_key(key)
            cmd = self._lookup(reason)
            while cmd is not None:
                reason, *cmdargs = cmd
                cmdstr = self._command_str(*cmdargs)
                print(f"\n{key!r} depends on {reason!r}: {cmdstr!r}")
                key = reason
                cmd = self._lookup(reason)
            print(f"\n{reason!r} is not defined\n")

    @property
    def ignored(self):
        """Set of ignored commands"""
        ignd = set()
        for cmd in self.postponed:
            if cmd[0] == self._gen_key(cmd[2]):
                ignd.add(cmd[2])
        return ignd

    def expand(self, key: str) -> Generator[elt.Element, None, None]:
        """iterator over AT objects generated by a source object"""
        self.force = True
        try:
            v = self[key]
            if isinstance(v, AnyDescr):
                yield from v.expand(self)
            else:
                yield v
        except Exception as exc:
            strargs = (arg for arg in exc.args if isinstance(arg, str))
            print(f"{type(exc).__name__}: {': '.join(strargs)}")
            self._analyse(self._reason(exc))
            raise
        finally:
            self.force = self.always_force

    def _generator(self, params):
        """Generate AT elements for the Lattice constructor"""
        use = params.setdefault("use", "ring")
        params.setdefault("name", use)
        params.setdefault("energy", 1.0e9)
        params.setdefault("periodicity", 1)

        # Iterate from the elements
        yield from self.expand(use)

    def lattice(self, use="ring", **kwargs):
        """Create a lattice from the selected sequence

        Parameters:
            use:                Name of the sequence or line containing the desired
              lattice. Default: ``ring``

        Keyword Args:
            name (str):         Name of the lattice. Default: sequence name.
            particle(Particle): Circulating particle. Default: Particle("relativistic")
            energy (float):     Energy of the lattice [eV]. Default: 1.0 GeV
            periodicity(int):   Number of periods. Default: 1
            *:                  All other keywords will be set as Lattice attributes
        """
        return Lattice(
            self._generator,
            iterator=params_filter,
            in_file=self.in_file or None,
            use=use,
            **kwargs,
        )

    def parse_lines(
        self,
        lines: Iterable[str],
        *,
        final: bool = True,
        **kwargs,
    ) -> None:
        """Process input lines and fill the database

        Args:
            lines: Iterable of input lines
            final: If :py:obj:`True`, signals that the undefined variables may be set
              to the default value
            **kwargs:   Initial variable definitions
        """
        self.update(**kwargs)
        buffer = []
        in_comment: bool = False
        ok: bool = True
        for line_number, contents in enumerate(lines):
            # Handle comments
            while contents:
                in_comment, contents = self.skip_comments(buffer, contents, in_comment)
            if contents is None:
                # Special case to avoid that empty lines break the continuation
                continue

            if buffer:
                contents = "".join(buffer).strip()
                buffer = []
                if not contents:
                    continue
            else:
                continue

            # Handle delimiters
            if self._delimiter is None:
                statements = []
                last = contents
            else:
                *statements, last = contents.split(sep=self._delimiter)

            # Handle continuation
            if self._continuation is None:
                buffer.append(last)
            else:
                idc = last.find(self._continuation)
                if idc >= 0:
                    buffer.append(last[:idc])
                else:
                    statements.append(last)

            # Process statements
            for stmnt in statements:
                stmnt = stmnt.strip()
                if not stmnt:
                    continue
                try:
                    ok = self._statement(stmnt)
                except Exception as exc:
                    message = f"Line {line_number} {stmnt!r}, {exc}"
                    print(message)
                    raise type(exc)(message) from exc
                if not ok:
                    break

            if not ok:
                break

        self._finalise(final=final)

    def parse_files(
        self,
        *filenames: str,
        final: bool = True,
        prolog: None | int | Callable[..., None] = None,
        epilog: Callable[..., None] | None = None,
        **kwargs,
    ) -> None:
        """Process files and fill the database

        Args:
            *filenames: Files to process
            final: If :py:obj:`True`, signals that the undefined variables may be set
              to the default value
            prolog:
            epilog:
            **kwargs:   Initial variable definitions
        """
        self.update(**kwargs)
        filenames = tuple(abspath(file) for file in filenames)
        self.in_file.extend(filenames)
        last = len(filenames) - 1
        ElementDescr._mentioned.clear()
        for nf, fn in enumerate(filenames):
            fn = normpath(join(self.bases[-1], fn))
            self.bases.append(dirname(fn))
            print("Processing", fn)
            try:
                with open(fn) as f:
                    if callable(prolog):
                        prolog(f)
                    elif isinstance(prolog, int):
                        for _i in range(prolog):
                            next(f)

                    self.parse_lines(f, final=final and (nf == last))

                    if callable(epilog):
                        epilog(f)
            finally:
                print("End", fn)
                self.bases.pop()


class UnorderedParser(BaseParser):
    """Parser allowing definitions in any order

    This is done by storing the failed statements in a queue and iteratively trying
    to execute them after all input statements have been processed, until the number
    of failures is constant (hopefully zero)
    """

    def __init__(self, env: dict, *args, **kwargs):
        """
        Args:
            env: global namespace
            verbose: If True, print detail on the processing
            *args: dict initializer
            **kwargs: dict initializer
        """
        super().__init__(env, *args, always_force=False, **kwargs)

    def _finalise(self, final: bool = True) -> None:
        """Loop on evaluation of the pending statements"""

        def replay():
            nend = len(self.postponed)
            while nend > 0:
                self._print(f"Delayed evaluation of {nend} statements")
                nstart = nend
                statements = self.postponed
                self.postponed = []
                for _reason, *args in statements:
                    self._decode(*args)
                nend = len(self.postponed)
                if nend == nstart:
                    break

        # at the end of each file: try again previously failing commands
        replay()

        # After the last file: initialize the remaining undefined variables
        default_value = self.get(self._undef_key)
        if final:
            undefined = self._missing(verbose=self.verbose)
            self._print(f"{len(undefined)} missing definitions.")
            if undefined and default_value is not None:
                for var in undefined:
                    self._print(f"Set {var} to {default_value}")
                    self[var] = default_value
                # last trial
                replay()


class LowerCaseParser(BaseParser):
    """Case independent parser"""

    @classmethod
    def _defkey(cls, expr: str, quoted: bool) -> str:
        """substitutions to get a valid python identifier"""
        expr = super()._defkey(expr, quoted)
        return expr if quoted else expr.lower()

    def _format_statement(self, line: str) -> str:
        """Reformat the input line"""
        return super()._format_statement(line.lower())


class UpperCaseParser(BaseParser):
    """Case independent parser"""

    @classmethod
    def _defkey(cls, expr: str, quoted: bool) -> str:
        """substitutions to get a valid python identifier"""
        expr = super()._defkey(expr, quoted)
        return expr if quoted else expr.upper()

    def _format_statement(self, line: str) -> str:
        """Reformat the input line"""
        return super()._format_statement(line.upper())



"""
Load lattices from Matlab files.
"""

from __future__ import annotations

__all__ = ["load_mat", "save_mat", "load_m", "save_m", "load_var"]

import sys
import os
from os.path import abspath, basename, splitext
from typing import Any
from collections.abc import Sequence, Generator
from math import isfinite
from warnings import warn

import numpy as np
import scipy.io

# imports necessary in 'globals()' for 'eval'
from numpy import array, uint8, nan as NaN  # noqa: F401

from .allfiles import register_format
from .utils import split_ignoring_parentheses, RingParam, keep_elements
from .utils import _drop_attrs, _CLASS_MAP
from ..lattice import Element, Lattice, Particle, Filter
from ..lattice import elements, AtWarning, params_filter, AtError

# Translation of RingParam attributes
_m2p = {
    "FamName": "name",
    "Energy": "energy",
    "Periodicity": "periodicity",
    "Particle": "particle",
    "cell_harmnumber": "cell_harmnumber",  # necessary: property
    "beam_current": "beam_current",  # necessary: property
    "PassMethod": None,  # Useless Matlab attributes
    "Length": None,
    "cavpts": None,
    "Mat_File": None,  # These are erroneous attributes saved in
    "Mat_Key": None,  # RingParam by old versions
    "Beam_Current": None,
    "Nbunch": None,
}
_p2m = {v: k for k, v in _m2p.items() if v is not None}
# Attribute to drop when writing a file
_p2m.update(_drop_attrs)

# Python to Matlab type translation
_mattype_map = {
    int: float,
    list: lambda attr: np.array(attr, dtype=object),
    tuple: lambda attr: np.array(attr, dtype=object),
    Particle: lambda attr: attr.to_dict(),
}
# Matlab constructor function
# Default: "".join(("at", element_class.__name__.lower()))
_mat_constructor = {
    "Dipole": "atsbend",
    "M66": "atM66",
}


def _mat_encoder(v):
    """type encoding for .mat files"""
    return _mattype_map.get(type(v), lambda attr: attr)(v)


def _matfile_generator(
    params: dict[str, Any], mat_file: str
) -> Generator[Element, None, None]:
    """Run through Matlab cells and generate AT elements"""

    def mclean(data):
        if data.dtype.type is np.str_:
            # Convert strings in arrays back to strings.
            return str(data[0]) if data.size > 0 else ""
        elif data.size == 1:
            v = data[0, 0]
            if issubclass(v.dtype.type, np.void):
                # Object => Return a dict
                return {f: mclean(v[f]) for f in v.dtype.fields}
            else:
                # Return a scalar
                return v
        else:
            # Remove any surplus dimensions in arrays.
            return np.squeeze(data)

    # noinspection PyUnresolvedReferences
    m = scipy.io.loadmat(params.setdefault("in_file", mat_file))
    matvars = [varname for varname in m if not varname.startswith("__")]
    default_key = matvars[0] if (len(matvars) == 1) else "RING"
    key = params.setdefault("use", default_key)
    if key not in m.keys():
        kok = [k for k in m.keys() if "__" not in k]
        raise AtError(
            f"Selected '{key}' variable does not exist, please select in: {kok}"
        )
    check = params.pop("check", True)
    quiet = params.pop("quiet", False)
    cell_array = m[key].flat
    for index, mat_elem in enumerate(cell_array):
        elem = mat_elem[0, 0]
        kwargs = {f: mclean(elem[f]) for f in elem.dtype.fields}
        yield Element.from_dict(kwargs, index=index, check=check, quiet=quiet)


def ringparam_filter(
    params: dict[str, Any], elem_iterator: Filter, *args
) -> Generator[Element, None, None]:
    """Run through all elements, process and optionally removes
    RingParam elements

    Parameters:
        params:         Lattice building parameters (see :py:class:`.Lattice`)
        elem_iterator:  Iterator over the lattice Elements

    Yields:
        elem (Element): new Elements

    The following keys in ``params`` are used:

    ============    ===================
    **keep_all**    keep RingParam elem_iterator as Markers
    ============    ===================

    The following keys in ``params`` are set:

    * ``name``
    * ``energy``
    * ``periodicity``
    * ``_harmnumber`` or
    * ``harmonic_number``
    * ``_particle``
    * ``_radiation``
    """
    keep_all = params.pop("keep_all", False)
    ringparams = []
    radiate = False
    for elem in elem_iterator(params, *args):
        if elem.PassMethod.endswith("RadPass") or elem.PassMethod.endswith(
            "CavityPass"
        ):
            radiate = True
        if isinstance(elem, RingParam):
            ringparams.append(elem)
            for k, v in elem.items():
                k2 = _m2p.get(k, k)
                if k2 is not None:
                    if k2 != "cell_harmnumber" or isfinite(v):
                        params.setdefault(k2, v)
            if keep_all:
                pars = vars(elem).copy()
                name = pars.pop("FamName")
                yield elements.Marker(name, tag="RingParam", **pars)
        else:
            yield elem
    params["_radiation"] = radiate
    params.setdefault("periodicity", 0)

    if len(ringparams) > 1:
        warn(
            AtWarning("More than 1 RingParam element, the 1st one is used"),
            stacklevel=2,
        )


def load_mat(filename: str, **kwargs) -> Lattice:
    """Create a :py:class:`.Lattice`  from a Matlab mat-file

    Parameters:
        filename:           Name of a '.mat' file

    Keyword Args:
        use (str):          Name of the Matlab variable containing
          the lattice. Default: it there is a single variable, use it, otherwise
          select 'RING'
        mat_key (str):      deprecated alias for *use*
        check (bool):       Run the coherence tests. Default:
          :py:obj:`True`
        quiet (bool):       Suppress the warning for non-standard
          classes. Default: :py:obj:`False`
        keep_all (bool):    Keep RingParam elements as Markers.
          Default: :py:obj:`False`
        name (str):         Name of the lattice. Default: taken from
          the lattice, or ``''``
        energy (float):     Energy of the lattice [eV]. Default: taken
          from the lattice elements
        periodicity(int):   Number of periods. Default: taken from the
          elements, or 1
        *:                  All other keywords will be set as Lattice
          attributes

    Returns:
        lattice (Lattice):  New :py:class:`.Lattice` object

    See Also:
        :py:func:`.load_lattice` for a generic lattice-loading function.
    """
    if "key" in kwargs:  # process the deprecated 'key' keyword
        kwargs.setdefault("use", kwargs.pop("key"))
    if "mat_key" in kwargs:  # process the deprecated 'mat_key' keyword
        kwargs.setdefault("use", kwargs.pop("mat_key"))
    return Lattice(
        ringparam_filter,
        _matfile_generator,
        abspath(filename),
        iterator=params_filter,
        **kwargs,
    )


def _element_from_m(line: str) -> Element:
    """Builds an :py:class:`.Element` from a line in an m-file

    Parameters:
        line:           Matlab string representation of an :py:class:`.Element`

    Returns:
        elem (Element): new :py:class:`.Element`
    """

    def argsplit(value):
        return [a.strip() for a in split_ignoring_parentheses(value)]

    def makedir(mat_struct):
        """Build directory from Matlab struct arguments"""

        def pairs(it):
            while True:
                try:
                    a = next(it)
                except StopIteration:
                    break
                yield eval(a), convert(next(it))

        return dict(pairs(iter(mat_struct)))

    def makearray(mat_arr):
        """Build numpy array for Matlab array syntax"""

        def arraystr(arr):
            lns = arr.split(";")
            rr = [arraystr(v) for v in lns] if len(lns) > 1 else lns[0].split()
            return f"[{', '.join(rr)}]"

        return eval(f"array({arraystr(mat_arr)})")

    def convert(value):
        """convert Matlab syntax to numpy syntax"""
        if value.startswith("["):
            result = makearray(value[1:-1])
        elif value.startswith("struct"):
            result = makedir(argsplit(value[7:-1]))
        else:
            result = eval(value)
        return result

    left = line.index("(")
    right = line.rindex(")")
    matcls = line[:left].strip()[2:]
    cls = _CLASS_MAP[matcls]
    arguments = argsplit(line[left + 1 : right])
    ll = len(cls._BUILD_ATTRIBUTES)
    if ll < len(arguments) and arguments[ll].endswith("Pass'"):
        arguments.insert(ll, "'PassMethod'")
    args = [convert(v) for v in arguments[:ll]]
    kwargs = makedir(arguments[ll:])
    if matcls == "rbend":
        # the Matlab 'rbend' has no equivalent in PyAT. This adds parameters
        # necessary for using the python sector bend
        halfangle = 0.5 * args[2]
        kwargs.setdefault("EntranceAngle", halfangle)
        kwargs.setdefault("ExitAngle", halfangle)
    return cls(*args, **kwargs)


def load_m(filename: str, **kwargs) -> Lattice:
    """Create a :py:class:`.Lattice`  from a Matlab m-file

    Parameters:
        filename:           Name of a '.m' file

    Keyword Args:
        keep_all (bool):    Keep RingParam elements as Markers.
          Default: :py:obj:`False`
        name (str):         Name of the lattice. Default: taken from
          the lattice, or ``''``
        energy (float):     Energy of the lattice [eV]. Default: taken
          from the lattice elements
        periodicity(int):   Number of periods. Default: taken from the
          elements, or 1
        *:                  All other keywords will be set as Lattice
          attributes

    Returns:
        lattice (Lattice):  New :py:class:`.Lattice` object

    See Also:
        :py:func:`.load_lattice` for a generic lattice-loading function.
    """

    def mfile_generator(params: dict, m_file: str) -> Generator[Element, None, None]:
        """Run through the lines of a Matlab m-file and generate AT elements"""
        with open(params.setdefault("in_file", m_file)) as file:
            _ = next(file)  # Matlab function definition
            _ = next(file)  # Cell array opening
            for lineno, line in enumerate(file):
                if line.startswith("};"):
                    break
                try:
                    elem = _element_from_m(line)
                except ValueError:
                    warn(AtWarning(f"Invalid line {lineno} skipped."), stacklevel=2)
                    continue
                except KeyError:
                    warn(AtWarning(f"Line {lineno}: Unknown class."), stacklevel=2)
                    continue
                else:
                    yield elem

    return Lattice(
        ringparam_filter,
        mfile_generator,
        abspath(filename),
        iterator=params_filter,
        **kwargs,
    )


def load_var(matlat: Sequence[dict], **kwargs) -> Lattice:
    """Create a :py:class:`.Lattice` from a Matlab cell array

    Parameters:
        matlat:             Matlab lattice

    Keyword Args:
        keep_all (bool):    Keep RingParam elements as Markers.
          Default: :py:obj:`False`
        name (str):         Name of the lattice. Default: taken from
          the lattice, or ``''``
        energy (float):     Energy of the lattice [eV]. Default: taken
          from the lattice elements
        periodicity(int):   Number of periods. Default: taken from the
          elements, or 1
        *:                  All other keywords will be set as Lattice
          attributes

    Returns:
        lattice (Lattice):  New :py:class:`.Lattice` object
    """

    # noinspection PyUnusedLocal
    def var_generator(params, latt):
        for elem in latt:
            yield Element.from_dict(elem)

    return Lattice(
        ringparam_filter, var_generator, matlat, iterator=params_filter, **kwargs
    )


def matlab_ring(ring: Lattice) -> Generator[Element, None, None]:
    """Prepend a RingParam element to a lattice"""

    def required(rng):
        # Public lattice attributes
        params = {k: v for k, v in vars(rng).items() if not k.startswith("_")}
        # Output the required attributes/properties
        for kp, km in _p2m.items():
            try:
                v = getattr(rng, kp)
            except AttributeError:
                pass
            else:
                params.pop(kp, None)
                if km is not None:
                    yield km, v
        # Output the remaining attributes
        yield from params.items()

    dct = dict(required(ring))
    yield RingParam(**dct)
    yield from keep_elements(ring)


def save_mat(ring: Lattice, filename: str, **kwargs) -> None:
    """Save a :py:class:`.Lattice` as a Matlab mat-file

    Parameters:
        ring:       Lattice description
        filename:   Name of the '.mat' file

    Keyword Args:
        use (str):  Name of the Matlab variable containing the lattice, Default: "RING"
        mat_key (str): Deprecated, alias for *use*

    See Also:
        :py:func:`.save_lattice` for a generic lattice-saving function.
    """
    # Ensure the lattice is a Matlab column vector: list(list)
    use = kwargs.pop("mat_key", "RING")  # For backward compatibility
    use = kwargs.pop("use", use)
    lring = [[el.to_dict(encoder=_mat_encoder)] for el in matlab_ring(ring)]
    scipy.io.savemat(filename, {use: lring}, long_field_names=True)


def _element_to_m(elem: Element) -> str:
    """Builds the Matlab-evaluable string for an :py:class:`.Element`

    Parameters:
        elem:           :py:class:`.Element`

    Returns:
        mstr (str):     Matlab string representation of the
          :py:class:`.Element` attributes
    """

    def convert(arg):
        def convert_dict(pdir):
            def scan(d):
                for k, v in d.items():
                    yield convert(k)
                    yield convert(v)

            return "struct({})".format(", ".join(scan(pdir)))

        def convert_array(arr):
            if arr.ndim > 1:
                return np.array2string(arg).replace("\n", ";")
            elif arr.ndim > 0:
                return np.array2string(arg)
            else:
                return str(arr)

        def convert_list(lst):
            return f"{{{{{str(lst)[1:-1]}}}}}"

        if isinstance(arg, np.ndarray):
            return convert_array(arg)
        elif isinstance(arg, np.number):
            return str(arg)
        elif isinstance(arg, dict):
            return convert_dict(arg)
        elif isinstance(arg, tuple):
            return convert_list(arg)
        elif isinstance(arg, list):
            return convert_list(arg)
        elif isinstance(arg, Particle):
            return convert_dict(arg.to_dict())
        else:
            return repr(arg)

    def m_name(elclass):
        classname = elclass.__name__
        return _mat_constructor.get(classname, "".join(("at", classname.lower())))

    attrs = dict(elem.items())
    # noinspection PyProtectedMember
    args = [attrs.pop(k, getattr(elem, k)) for k in elem._BUILD_ATTRIBUTES]
    defelem = elem.__class__(*args)
    kwds = {
        k: v
        for k, v in attrs.items()
        if not np.array_equal(v, getattr(defelem, k, None))
    }
    argstrs = [convert(arg) for arg in args]
    if "PassMethod" in kwds:
        argstrs.append(convert(kwds.pop("PassMethod")))
    argstrs += [", ".join((repr(k), convert(v))) for k, v in kwds.items()]
    return "{:>15}({});...".format(m_name(elem.__class__), ", ".join(argstrs))


def save_m(ring: Lattice, filename: str | None = None) -> None:
    """Save a :py:class:`.Lattice` as a Matlab m-file

    Parameters:
        ring:           Lattice description
        filename:       Name of the '.m' file. Default: outputs on
          :py:obj:`sys.stdout`

    See Also:
        :py:func:`.save_lattice` for a generic lattice-saving function.
    """

    def save(file):
        with np.printoptions(linewidth=1000, floatmode="unique"):
            print("ring = {...", file=file)
            for elem in matlab_ring(ring):
                print(_element_to_m(elem), file=file)
            print("};", file=file)

    if filename is None:
        save(sys.stdout)
    else:
        with open(filename, "w") as mfile:
            [funcname, _] = splitext(basename(filename))
            print(f"function ring = {funcname}()", file=mfile)
            save(mfile)
            print("    function v=False()\n        v=false;\n    end", file=mfile)
            print("    function v=True()\n        v=true;\n    end", file=mfile)
            print("end", file=mfile)


# Simulates the deprecated "mat_file" and "mat_key" attributes
def _mat_file(ring):
    """.mat input file. Deprecated, use *in_file* instead."""
    try:
        in_file = ring.in_file
    except AttributeError:
        raise AttributeError("'Lattice' object has no attribute 'mat_file'") from None
    if isinstance(in_file, str):
        _, ext = os.path.splitext(in_file)
        if ext != ".mat":
            raise AttributeError("'Lattice' object has no attribute 'mat_file'")
    else:
        raise AttributeError("'Lattice' object has no attribute 'mat_file'")
    return in_file


def _mat_key(ring):
    """selected Matlab variable. Deprecated, use *use* instead."""
    try:
        mat_key = ring.use
    except AttributeError:
        raise AttributeError("'Lattice' object has no attribute 'mat_key'") from None
    return mat_key


# noinspection PyUnusedLocal
def _ignore(ring, value):
    pass


register_format(
    ".mat",
    load_mat,
    save_mat,
    descr="Matlab binary mat-file. See :py:func:`.load_mat`.",
)

register_format(
    ".m",
    load_m,
    save_m,
    descr="Matlab text m-file. See :py:func:`.load_m`.",
)

Lattice.mat_file = property(_mat_file, _ignore, None)
Lattice.mat_key = property(_mat_key, _ignore, None)



"""AT plotting functions related to resonances."""

from __future__ import annotations

__all__ = ["farey_sequence", "plot_tune_diagram", "create_linepalette"]

import warnings
from fractions import Fraction

import matplotlib.axes
import matplotlib.pyplot as plt
import numpy as np

# 2024jul31 oblanco at ALBA CELLS


def create_linepalette(
    linestyle: str | dict = None,
    linecolor: str = None,
    linewidth: int = None,
    addtolabel: str = None,
) -> dict[str, any]:
    """Create a line palette to plot resonance lines.

    Parameters:
        linestyle: If a dictionary is passed, it should contain
            {"normal": style1, "skew": style2}

            If 'dots' it uses dotted styles as linestyles. Equivalent to:
            {"normal": "dashdot", "skew": "dotted"}

            Default: {"normal": '-', "skew": '--'}
        linecolor: line color, e.g. "k". Default: custom values.
            See :py:func:`plot_tune_diagram`
        linewidth: line width. Default: custom values. See :py:func:`plot_tune_diagram`
        addtolabel: adds a string to the line label

    Returns:
        palette: dictionary containing the line properties for resonance plots.
    """
    # create default dictionary with line properties
    mypalettecolor1 = {
        1: "k",
        2: "b",
        3: "r",
        4: "g",
        5: "m",
        6: "c",
        7: "y",
        8: "darkcyan",
        9: "lightgreen",
        10: (0.1, 0.1, 0.1),
        11: (0.1, 0.1, 0.1),
        12: (0.2, 0.2, 0.2),
        13: (0.3, 0.3, 0.3),
        14: (0.4, 0.4, 0.4),
        15: (0.5, 0.5, 0.5),
    }
    mypalettewidth_main = {
        1: 4,
        2: 3,
        3: 2,
        4: 1,
        5: 1,
        6: 1,
        7: 1,
        8: 1,
        9: 1,
        10: 1,
        11: 1,
        12: 1,
        13: 1,
        14: 1,
        15: 1,
    }
    mypalettestyle_main = {"normal": "-", "skew": "--"}
    mypalettestyle_withdots = {"normal": "dashdot", "skew": "dotted"}
    # set the line style
    mypalettestyle = mypalettestyle_main
    if linestyle == "dots":
        mypalettestyle = mypalettestyle_withdots
    if isinstance(linestyle, dict):
        mypalettestyle = linestyle
    # set the line color
    mypalettecolor = mypalettecolor1
    if linecolor is not None:
        for lsize in range(1, 16):
            mypalettecolor[lsize] = linecolor
    # set the line width
    mylinewidth = mypalettewidth_main
    if isinstance(linewidth, int) and linewidth > 0:
        for lsize in range(1, 16):
            mylinewidth[lsize] = linewidth
    # add to label
    if addtolabel is None:
        addtolabel = ""
    # set the dictionary
    linepropdict = {}
    resontypes = ["normal", "skew"]
    for resontype in resontypes:
        linepropdict[resontype] = {}
    for resontype in resontypes:
        for idx in range(1, 16):
            propaux = {
                "color": mypalettecolor[idx],
                "linestyle": mypalettestyle[resontype],
                "linewidth": mylinewidth[idx],
                "label": str(idx) + resontype[0] + addtolabel,
            }
            linepropdict[resontype][idx] = propaux
    return linepropdict


def farey_sequence(nthorder: int, verbose: bool = False) -> tuple[list, list]:
    """
    Return the Farey sequence, and the resonance sequence of nth order.

    Parameters:
        nthorder: natural number bigger than 0.
        verbose: prints extra info. Default: False.

    Returns:
        fareyseqfloat: list of elements with the Farey sequence in Float format.
            See Eqs.(1,2,3) of [1].
        fareyseqfrac: list of elements with the Farey sequence in Fraction format.
            See Eqs.(1,2,3) of [1].

    Raises:
        ValueError: if given order is lower than 0, or window is zero.

    [1] R.Tomas. 'From Farey sequences to resonance diagrams.
            Phys.Rev.Acc.Beams 17, 014001 (2014)'
    """
    # verboseprint to check flag only once
    verboseprint = print if verbose else lambda *a, **k: None

    if nthorder < 1:
        raise ValueError("The order must be a positive integer larger than zero.")
    verboseprint(f"nthorder={nthorder}")

    afarey = 0
    bfarey = 1
    cfarey = 1
    dfarey = nthorder
    farey = [0, 1 / dfarey]
    fracfarey = [Fraction(0), Fraction(1, dfarey)]
    idx = 0
    while (farey[-1] < 1) and (idx < 100):
        idx += 1
        caux = np.floor((nthorder + bfarey) / dfarey) * cfarey - afarey
        daux = np.floor((nthorder + bfarey) / dfarey) * dfarey - bfarey
        afarey = cfarey
        bfarey = dfarey
        cfarey = int(caux)
        dfarey = int(daux)
        farey.append(cfarey / dfarey)
        fracfarey.append(Fraction(cfarey, dfarey))
    verboseprint(f"farey_float{nthorder}= {farey}")
    verboseprint(f"farey_frac{nthorder} = {fracfarey}")
    return farey, fracfarey


def plot_tune_diagram(
    orders: int | tuple[int] = (1, 2, 3),
    periodicity: int = 1,
    window: list = (0, 1, 0, 1),
    verbose: bool = False,
    legend: bool = False,
    show: bool = True,
    block: bool = False,
    debug: bool = False,
    axes: matplotlib.axes.Axes = None,
    linestyle: dict or str = None,
    linecolor: str or any = None,
    linewidth: int = None,
    addtolabel: str = None,
    **kwargs: dict[str, any],
) -> tuple[matplotlib.axes.Axes, list, list]:
    r"""
    Plot the tune diagram and resonance lines for the given *orders*, *periodicity*
    and *window*.

    The resonance equation is :math:`a\nu_x + b\nu_y = c`
    with :math:`a,b` and :math:`c` integers. The order is: :math:`N=abs(a)+abs(b)`.

    Parameters:
        orders: integer or tuple of integers larger than zero. Default (1, 2, 3).
        periodicity: periodicity of the machine, integer larger than zero. Default: 1.
        window: ``(nux_min, nux_max, nuy_min, nuy_max)``: tuple of 4 values for the
            tune minimum and maximum window. Default: (0, 1, 0, 1).
            *window* is ignored if the parameter axes is given.
        verbose: print verbose output.
        legend: print legend on the plot. Default: False.
        show: show plot. Default: True.
        block: passed to plot.show(). Default: False.
        debug: extra output to check line construction. Default: False.
        axes: :py:class:`~matplotlib.axes.Axes` for plotting the
            resonances. If :py:obj:`None`, a new figure will be created.
            Note that if *axes* are given then *window* is ignored.
        linestyle: line style for normal and skew resonances.

            If a dictionary is passed, it should contain
            {"normal": style1, "skew": style2}

            If 'dots' it uses dotted styles as linestyles. Equivalent to:
            {"normal": "dashdot", "skew": "dotted"}

            Default: {"normal": '-', "skew": '--'}
        linecolor: single color for all resonances. Default: custom palette.
            See :ref:`Lines Color and Width <color_width>`
        linewidth: line width for all resonances. Default: custom values.
            See :ref:`Lines Color and Width <color_width>`
        addtolabel: adds a string to the line label, e.g. for the fourth
            order normal resonance "4n"+addtolabel
    Keyword Args:
        only (str): if 'normal', plot only normal resonances.

            if 'skew' plots only skew resonances.

            Otherwise, ignored. See the notes on Normal and Skew convention.
        linedict (dict): dictionary of custom line styles. See notes below.

    Returns:
        Axes (matplotlib.axes.Axes): object from matplotlib.axes._axes
        legend_h (list):  list of handles for the legend
        legend_lab (list): list of labels for the legend

    NOTES:

    Normal and Skew convention:
    Line style is similar to reson.m from Matlab Middle Layer, MML, by L. Nadolski.
    Normal resonances are plotted with a continuous line.
    Skew resonances, i.e. N-abs(a) is odd, are plotted in dashed lines.

    .. _color_width:

    Lines Color and Width:
    Line style is similar to reson.m from Matlab Middle Layer, MML, by L. Nadolski.
    1st: black, width 4
    2nd: blue, width 3
    3rd: red, width 2
    4th: green, width 1
    5th: magenta, witdh 1
    6th: cyan, width 1
    7th: yellow, width 1
    8th: darkcyan, width 1
    9th: 'lightgreen', width 1
    10th to 15th: RGB increased in steps of 0.1, width 1

    Custom Style: You could pass a custom line style in a dictionary as
    ``linedict=mydictionary``, where mydictionary should contain two entries:
    dict("normal": normald, "skew": skewd).
    normald and skewd are also dictionaries, each entry contains as key
    the resonance order and as value the line properties to use in the plot.
    The default dictionary is created with :py:func:`create_linepalette`
    mydictionary = at.plot.resonances.create_linepalette()
    you could edit the needed entries.

    Raises:
        ValueError: if given resonances are lower than 0, or window is zero.
    """
    # verboseprint to check flag only once
    verboseprint = print if verbose else lambda *a, **k: None

    # print debugging output, equivalent to extra verbose
    debugprint = print if debug else lambda *a, **k: None

    # only plot normal or skew
    normalskew = [0, 1]
    onlyns = kwargs.pop("only", False)
    if onlyns == "normal":
        normalskew = [0]
    if onlyns == "skew":
        normalskew = [1]

    # orders could be a single int
    if isinstance(orders, int):
        orders = [orders]
    # check that all are larger than 0
    if sum(n < 1 for n in orders):
        raise ValueError("Negative resonances are not allowed.")

    if periodicity < 1:
        raise ValueError("Period must be a positive integer.")
    verboseprint(f"The periodicity is {periodicity}")

    verboseprint(f"The window is {window}")
    # check the window
    windowa = np.array(window)
    if windowa[0] == windowa[1]:
        raise ValueError("horizontal coordinates must be different")
    if windowa[2] == windowa[3]:
        raise ValueError("vertical coordinates must be different")
    if windowa[1] < windowa[0]:
        windowa[0], windowa[1] = windowa[1], windowa[0]
        warnings.warn("Swapping horizontal coordinates", stacklevel=1)
    if windowa[3] < windowa[2]:
        windowa[2], windowa[3] = windowa[3], windowa[2]
        warnings.warn("Swapping vertical coordinates", stacklevel=1)
    # get xlimits and ylimits
    the_axeslims = windowa.reshape((2, 2))

    maxreson2calc = np.max(orders)
    verboseprint(f"Farey max order={maxreson2calc}")

    # get the Farey collection, i.e., a list of farey sequences, one per order
    fareycollectionfloat = {}
    fareycollectionfrac = {}
    for nthorder in range(1, maxreson2calc + 1):
        farey, fracfarey = farey_sequence(nthorder)
        fareycollectionfloat[nthorder] = farey
        fareycollectionfrac[nthorder] = fracfarey
    verboseprint(f"the Farey collection is {fareycollectionfloat}")

    # plot configuration
    if axes is None:
        fig = plt.figure()
        axes = fig.add_subplot(111)
    else:
        verboseprint("Axes already exist, ignore window")
        the_axeslims = np.array([axes.get_xlim(), axes.get_ylim()])
    # min/max to plot lines with slopes
    minx = np.floor(the_axeslims[0, 0])
    minx = minx - periodicity - np.mod(minx, periodicity)
    maxx = np.ceil(the_axeslims[0, 1])
    maxx = maxx + periodicity - np.mod(maxx, periodicity)
    minmaxxdist = maxx - minx
    miny = np.floor(the_axeslims[1, 0])
    miny = miny - periodicity - np.mod(miny, periodicity)
    maxy = np.ceil(the_axeslims[1, 1])
    maxy = maxy + periodicity - np.mod(maxy, periodicity)
    minmaxydist = maxy - miny

    lprop = create_linepalette(
        linestyle=linestyle,
        linecolor=linecolor,
        linewidth=linewidth,
        addtolabel=addtolabel,
    )
    userprop = kwargs.pop("linedict", {})
    lprop["normal"].update(userprop.get("normal", {}))
    lprop["skew"].update(userprop.get("skew", {}))

    # we only need two points to define a line
    nauxpoints = 2

    # window min/max,horizontal and vertical
    axes.set_xlim(the_axeslims[0, :])
    axes.set_ylim(the_axeslims[1, :])
    # start to check the Farey collection, starting with 0
    collectaux1 = [0]
    idxtotype = {0: "normal", 1: "skew"}
    for nthorder in range(1, maxreson2calc + 1):
        debugprint(f"nthorder={nthorder}")
        collectaux2 = fareycollectionfloat[nthorder]
        thesteps = list(set(collectaux2) - set(collectaux1))
        debugprint(thesteps)
        chosenstep = min(thesteps)
        debugprint(f"chosenstep={chosenstep}")
        collectaux1 = collectaux2
        if nthorder in orders:
            # increase step by the periodicity in straight resonances
            debugprint("enter plotting horizontal straight lines")
            if 0 in normalskew:
                for iaux in np.arange(minx, maxx + 0.000001, periodicity * chosenstep):
                    axes.axvline(x=iaux, **lprop["normal"][nthorder])
            debugprint("enter plotting vertical straight lines")
            nsaux = np.mod(nthorder, 2)
            if nsaux in normalskew:
                for iaux in np.arange(miny, maxy + 0.000001, periodicity * chosenstep):
                    axes.axhline(y=iaux, **lprop[idxtotype[nsaux]][nthorder])
            # aeq*nux + beq*nuy = nthorder
            for aeq in range(1, nthorder):
                debugprint(f"enter plotting diagonals {aeq}")
                beq = nthorder - aeq
                chosenslope = 1.0 * aeq / beq
                ystep = 1.0 / beq
                debugprint(f"chosen slope={chosenslope}")
                debugprint(f"chosen ystep={ystep}")
                # calculate the variation on the vertical direction from window size
                yyaux = np.ceil(minmaxxdist * chosenslope + minmaxydist * ystep)
                y1aux = -yyaux + miny
                y2aux = yyaux + maxy
                # adapt to periodicity
                y1aux = periodicity * np.floor(y1aux / periodicity)
                y2aux = periodicity * np.ceil(y2aux / periodicity)
                debugprint(f"minx={minx},maxx={maxx},minmaxxdist={minmaxxdist}")
                debugprint(f"miny={miny},maxy={maxy},minmaxydist={minmaxydist}")
                debugprint(f"y1aux={y1aux},y2aux={y2aux}")
                xaux = np.linspace(minx, maxx, nauxpoints)
                debugprint(f"xaux={xaux}")
                nsaux = np.mod(beq, 2)
                if nsaux in normalskew:
                    for istep in np.arange(0, y2aux - y1aux + 0.0001, ystep):
                        y1line = (
                            -chosenslope * (xaux - minx) + y2aux - periodicity * istep
                        )
                        y2line = (
                            chosenslope * (xaux - minx) + y1aux + periodicity * istep
                        )
                        debugprint(f"y1line={y1line},y2line={y2line}")
                        axes.plot(xaux, y1line, **lprop[idxtotype[nsaux]][nthorder])
                        axes.plot(xaux, y2line, **lprop[idxtotype[nsaux]][nthorder])
    # include labels
    axes.set_xlabel(r"$\nu_x$")
    axes.set_ylabel(r"$\nu_y$")
    # printing legend if necessary
    handleall, labelall = axes.get_legend_handles_labels()
    hstyle = [hline._linestyle for hline in handleall]
    joinlabelstyle = [i + j for i, j in zip(labelall, hstyle)]
    uniquelabelstyle = sorted(set(joinlabelstyle))
    idxunique = [joinlabelstyle.index(uniqele) for uniqele in uniquelabelstyle]
    myleghandles = [handleall[idx] for idx in idxunique]
    myleglabels = [labelall[idx] for idx in idxunique]
    if legend:
        axes.legend(handles=myleghandles, labels=myleglabels, frameon=legend)
    if show:
        plt.show(block=block)

    return axes, myleghandles, myleglabels



r"""Grouping of :py:class:`.Observable` objects for fast evaluation.

:py:class:`ObservableList`\ (lattice, ...)
    This container based on :py:class:`list` is in charge of optics
    computations and provides each individual :py:class:`.Observable` with its
    specific data.

:py:class:`ObservableList` provides the :py:meth:`~ObservableList.evaluate`
method and the following properties:

- :py:attr:`~ObservableList.values`
- :py:attr:`~ObservableList.flat_values`
- :py:attr:`~ObservableList.weights`
- :py:attr:`~ObservableList.flat_weights`
- :py:attr:`~ObservableList.weighted_values`
- :py:attr:`~ObservableList.flat_weighted_values`
- :py:attr:`~ObservableList.deviations`
- :py:attr:`~ObservableList.flat_deviations`
- :py:attr:`~ObservableList.residuals`
- :py:attr:`~ObservableList.sum_residuals`
"""

from __future__ import annotations

__all__ = [
    "ObservableList",
]

from collections.abc import Iterable, Iterator
from functools import reduce
from typing import Callable

import numpy as np
import numpy.typing as npt

# noinspection PyProtectedMember
from .observables import Observable, ElementObservable, Need
from .rdt_observable import RDTObservable
from ..lattice import AtError, frequency_control
from ..lattice import Lattice, Orbit, Refpts, All
from ..physics import linopt6
from ..tracking import internal_lpass


def _flatten(vals, order="F") -> npt.NDArray[float]:
    return np.concatenate([np.reshape(v, -1, order=order) for v in vals])


class _ObsResIter(Iterator):
    """Iterator object for the _ObsResult class"""

    def __init__(self, obsiter):
        self.base = obsiter

    def __next__(self):
        # Raises the stored error when reaching a missing value
        return Observable.check_value(next(self.base))


class _ObsResults(tuple):
    """Tuple-like object for the output of  ObservableList.evaluate

    _ObsResult implements a special treatment when the evaluation ends with an error.
    The error is stored instead of the value. This object raises the error a posteriori,
    when accessing the missing value.
    """

    def __getitem__(self, item):
        # Raises the stored error when accessing a missing value
        if isinstance(item, slice):
            return _ObsResults(super().__getitem__(item))
        else:
            return Observable.check_value(super().__getitem__(item))

    def __iter__(self):
        return _ObsResIter(super().__iter__())

    def __repr__(self):
        return repr(tuple(super().__iter__()))


class ObservableList(list):
    """Handles a list of Observables to be evaluated together.

    :py:class:`ObservableList` supports all :py:class:`list` methods, like
    appending, insertion or concatenation with the "+" operator.
    """

    needs_ring = {
        Need.RING,
        Need.ORBIT,
        Need.MATRIX,
        Need.GLOBALOPTICS,
        Need.LOCALOPTICS,
        Need.TRAJECTORY,
        Need.EMITTANCE,
        Need.GEOMETRY,
    }
    needs_orbit = {
        Need.ORBIT,
        Need.MATRIX,
        Need.GLOBALOPTICS,
        Need.LOCALOPTICS,
        Need.EMITTANCE,
    }
    needs_optics = {Need.GLOBALOPTICS, Need.LOCALOPTICS}

    def __init__(
        self,
        obsiter: Iterable[Observable] = (),
        *,
        method: Callable = linopt6,
        orbit: Orbit = None,
        twiss_in=None,
        r_in: Orbit = None,
        **kwargs,
    ):
        # noinspection PyUnresolvedReferences
        r"""
        Args:
            obsiter:    Iterable of :py:class:`.Observable`\ s

        Keyword Args:
            orbit (Orbit):      Initial orbit. Avoids looking for the closed
              orbit if it is already known. Used for
              :py:class:`MatrixObservable` and :py:class:`LocalOpticsObservable`
            twiss_in:           Initial conditions for transfer line optics.
              See :py:func:`.get_optics`. Used for
              :py:class:`LocalOpticsObservable`
            method (Callable):  Method for linear optics. Used for
              :py:class:`LocalOpticsObservable`.
              Default: :py:obj:`~.linear.linopt6`
            r_in (Orbit):       Initial trajectory, used for
              :py:class:`TrajectoryObservable`, Default: zeros(6)


        Example:
            >>> obslist = ObservableList()

            Create an empty Observable list

            >>> obslist.append(OrbitObservable(at.Monitor, plane="x"))
            >>> obslist.append(GlobalOpticsObservable("tune"))
            >>> obslist.append(EmittanceObservable("emittances", plane="h"))

            Add observables for horizontal closed orbit at Monitor locations,
            tunes and horizontal emittance

            >>> obslist.evaluate(ring, initial=True)

            Compute the values and mark them as the initial value

            >>> obslist.values
            [array([-3.02189464e-09,  4.50695130e-07,  4.08205818e-07,
                     2.37899969e-08, -1.31783561e-08,  2.47230794e-08,
                     2.95310770e-08, -4.05598110e-07, -4.47398092e-07,
                     2.24850671e-09]),
            array([3.81563019e-01, 8.54376397e-01, 1.09060730e-04]),
            1.320391045951568e-10]

            >>> obslist.get_flat_values("tune", "emittances[h]")
            array([3.815630e-01, 8.543764e-01, 1.090607e-04, 1.320391e-10])

            Get a flattened array of tunes and horizontal emittance
        """
        self.orbitrefs = None
        self.opticsrefs = None
        self.passrefs = None
        self.matrixrefs = None
        self.rdtrefs = None
        self.needs = None
        self.rdt_type = set()
        self.method = method
        self.orbit = orbit
        self.twiss_in = twiss_in
        self.r_in = r_in
        self.kwargs = kwargs
        super().__init__(obsiter)

    # noinspection PyProtectedMember
    def _setup(self, ring: Lattice):
        # Compute the union of all needs
        needs = set()
        rdt_type = set()
        for obs in self:
            needs |= obs.needs
            if isinstance(obs, ElementObservable):
                needs.add(Need.RING)
            if isinstance(obs, RDTObservable):
                rdt_type.add(obs._rdt_type)

        if (needs & self.needs_ring) and ring is None:
            raise ValueError("At least one Observable needs a ring argument")

        self.needs = needs
        self.rdt_type = rdt_type
        if ring is None:
            # Initialise each observable
            for obs in self:
                obs._setup(ring)
        else:
            # Initialise each observable and make a summary all refpoints
            noref = ring.get_bool_index(None)
            orbitrefs = opticsrefs = passrefs = matrixrefs = rdtrefs = noref
            for obs in self:
                obs._setup(ring)
                obsneeds = obs.needs
                if isinstance(obs, ElementObservable):
                    if Need.ORBIT in obsneeds:
                        orbitrefs |= obs._boolrefs
                    if Need.MATRIX in obsneeds:
                        matrixrefs |= obs._boolrefs
                    if Need.LOCALOPTICS in obsneeds:
                        if Need.ALL_POINTS in obsneeds:
                            opticsrefs = ring.get_bool_index(All)
                        else:
                            opticsrefs |= obs._boolrefs
                    if Need.TRAJECTORY in obsneeds:
                        passrefs |= obs._boolrefs
                    if Need.RDT in obsneeds:
                        rdtrefs |= obs._boolrefs
            self.orbitrefs = orbitrefs
            self.opticsrefs = opticsrefs
            self.rdtrefs = rdtrefs
            self.passrefs = passrefs
            self.matrixrefs = matrixrefs

    def __iadd__(self, other: ObservableList):
        if not isinstance(other, ObservableList):
            raise TypeError(f"Cannot add a {type(other)} to an ObservableList")
        self.extend(other)
        return self

    def __add__(self, other) -> ObservableList:
        nobs = ObservableList(self)
        nobs += other
        return nobs

    def append(self, obs: Observable):
        """Append observable to the end of the list."""
        if not isinstance(obs, Observable):
            raise TypeError(f"Cannot append a {type(obs)} to an ObservableList")
        self.needs = None
        super().append(obs)

    def extend(self, obsiter: Iterable[Observable]):
        """Extend list by appending Observables from the iterable."""
        self.needs = None
        super().extend(obsiter)

    def insert(self, index: int, obs: Observable):
        """Insert Observable before index."""
        if not isinstance(obs, Observable):
            raise TypeError(f"Cannot insert a {type(obs)} in an ObservableList")
        self.needs = None
        super().insert(index, obs)

    # noinspection PyProtectedMember
    def __str__(self):
        values = "\n".join(obs._all_lines() for obs in self)
        return "\n".join((Observable._header(), values))

    def evaluate(
        self,
        ring: Lattice | None = None,
        *,
        dp: float | None = None,
        dct: float | None = None,
        df: float | None = None,
        initial: bool = False,
        **kwargs,
    ):
        r"""Compute all the :py:class:`Observable` values.

        Args:
            ring:           Lattice used for evaluation
            dp (float):     Momentum deviation. Defaults to :py:obj:`None`
            dct (float):    Path lengthening. Defaults to :py:obj:`None`
            df (float):     Deviation from the nominal RF frequency.
              Defaults to :py:obj:`None`
            initial:    If :py:obj:`True`, store the values as *initial values*

        Keyword Args:
            orbit (Orbit):  Initial orbit. Avoids looking for the closed
              orbit if it is already known. Used for
              :py:class:`.MatrixObservable` and :py:class:`.LocalOpticsObservable`
            twiss_in:       Initial conditions for transfer line optics.
              See :py:func:`.get_optics`. Used for
              :py:class:`.LocalOpticsObservable`
            method (Callable):  Method for linear optics. Used for
              :py:class:`.LocalOpticsObservable`.
              Default: :py:obj:`~.linear.linopt6`
            r_in (Orbit):   Initial trajectory, used for
              :py:class:`.TrajectoryObservable`, Default: zeros(6)
        """

        def obseval(ring, obs):
            """Evaluate a single observable."""

            def check_error(data, refpts):
                return data if isinstance(data, Exception) else data[refpts]

            obsneeds = obs.needs
            obsrefs = getattr(obs, "_boolrefs", None)
            data = []
            if Need.RING in obsneeds:
                data.append(ring)
            if Need.ORBIT in obsneeds:
                data.append(check_error(orbits, obsrefs[self.orbitrefs]))
            if Need.MATRIX in obsneeds:
                data.append(check_error(mxdata, obsrefs[self.matrixrefs]))
            if Need.GLOBALOPTICS in obsneeds:
                data.append(rgdata)
            if Need.LOCALOPTICS in obsneeds:
                data.append(check_error(eldata, obsrefs[self.opticsrefs]))
            if Need.TRAJECTORY in obsneeds:
                data.append(trajs[obsrefs[self.passrefs]])
            if Need.EMITTANCE in obsneeds:
                data.append(emdata)
            if Need.GEOMETRY in obsneeds:
                data.append(geodata[obsrefs])
            if Need.RDT in obsneeds:
                data.append(check_error(rdtdata, obsrefs[self.rdtrefs]))
            return obs.evaluate(*data, initial=initial)

        @frequency_control
        def ringeval(
            ring,
            dp: float | None = None,
            dct: float | None = None,
            df: float | None = None,
        ):
            """Optics computations."""
            keep_lattice = False
            trajs = orbits = rgdata = eldata = emdata = mxdata = geodata = rdtdata = (
                None
            )
            twiss_in = kwargs.get("twiss_in", self.twiss_in)
            o0 = kwargs.get("orbit", self.orbit)
            o0 = getattr(twiss_in, "closed_orbit", None) if o0 is None else o0
            needs = self.needs
            needs_o0 = (needs & self.needs_orbit) and (o0 is None)

            if Need.TRAJECTORY in needs:
                # Trajectory computation
                r_in = kwargs.get("r_in", self.r_in)
                if r_in is None:
                    r_in = np.zeros(6)
                r_out = internal_lpass(ring, r_in.copy(), 1, refpts=self.passrefs)
                trajs = r_out[:, 0, :, 0].T
                keep_lattice = True

            # if needs & self.needs_orbit:
            if Need.ORBIT in needs or needs_o0:
                # Closed orbit computation
                try:
                    o0, orbits = ring.find_orbit(
                        refpts=self.orbitrefs,
                        dp=dp,
                        dct=dct,
                        df=df,
                        orbit=o0,
                        keep_lattice=keep_lattice,
                    )
                except AtError as err:
                    orbits = mxdata = rgdata = eldata = emdata = err
                else:
                    keep_lattice = True

            if Need.MATRIX in needs and o0 is not None:
                # Transfer matrix computation
                find_m = ring.find_m66 if ring.is_6d else ring.find_m44
                # noinspection PyUnboundLocalVariable
                _, mxdata = find_m(
                    refpts=self.matrixrefs,
                    dp=dp,
                    dct=dct,
                    df=df,
                    orbit=o0,
                    keep_lattice=keep_lattice,
                )
                keep_lattice = True

            if (needs & self.needs_optics) and o0 is not None:
                # Linear optics computation
                try:
                    _, rgdata, eldata = ring.get_optics(
                        refpts=self.opticsrefs,
                        dp=dp,
                        dct=dct,
                        df=df,
                        orbit=o0,
                        keep_lattice=keep_lattice,
                        get_chrom=Need.CHROMATICITY in needs,
                        get_w=Need.W_FUNCTIONS in needs,
                        twiss_in=twiss_in,
                        method=kwargs.get("method", self.method),
                    )
                except AtError as err:
                    rgdata = eldata = err
                else:
                    keep_lattice = True

            if Need.EMITTANCE in needs and o0 is not None:
                # Emittance computation
                try:
                    emdata = ring.envelope_parameters(
                        orbit=o0, keep_lattice=keep_lattice
                    )
                except Exception as err:
                    emdata = err

            if Need.GEOMETRY in needs:
                # Geometry computation
                geodata, _ = ring.get_geometry()

            if Need.RDT in needs:
                # RDT computation
                use_mp = kwargs.get("use_mp", False)
                pool_size = kwargs.get("pool_size", None)
                try:
                    _, _, rdtdata = ring.get_rdts(
                        refpts=self.rdtrefs,
                        rdt_type=self.rdt_type,
                        second_order=Need.RDT_2ND_ORDER in needs,
                        use_mp=use_mp,
                        pool_size=pool_size,
                    )
                except Exception as err:
                    rdtdata = err

            return trajs, orbits, rgdata, eldata, emdata, mxdata, geodata, rdtdata

        if self.needs is None or initial:
            self._setup(ring)

        trajs, orbits, rgdata, eldata, emdata, mxdata, geodata, rdtdata = ringeval(
            ring, dp=dp, dct=dct, df=df
        )
        return _ObsResults(obseval(ring, ob) for ob in self)

    def check(self) -> bool:
        """Check if all observables are evaluated.

        Returns:
            ok: :py:obj:`True` if evaluation is done, :py:obj:`False` otherwise

        Raises:
            AtError:    any value is doubtful: evaluation failed, empty value…
        """
        return all(obs.check() for obs in self)

    # noinspection PyProtectedMember
    def exclude(self, obsname: str, excluded: Refpts):
        """Set the excluded mask on the selected observable."""
        for obs in self:
            if obs.name == obsname:
                obs._excluded = excluded
        self.needs = None

    def _lookup(self, *ids: int | str) -> list[Observable]:
        """Observable lookup function"""

        def select(id):
            if isinstance(id, str):
                for obs in self:
                    if obs.name == id:
                        return obs
                else:
                    raise KeyError(id)
            else:
                return self[id]

        if ids:
            return [select(id) for id in ids]
        else:
            return self

    def _collect(self, attrname: str, *obsid: str | int, err: float | None = None):
        def val(obs):
            try:
                res = getattr(obs, attrname)
            except AtError:
                if err is None:
                    raise
                else:
                    # noinspection PyProtectedMember
                    shp = obs._shape
                    res = err if shp is None else np.broadcast_to(err, shp)
            return res

        obslist = self._lookup(*obsid)
        return tuple(val(obs) for obs in obslist)

    def get_shapes(self, *obsid: str | int) -> tuple:
        """Return the shapes of all values.

        Args:
            *obsid: name or index of selected observables (Default all)
        """
        return self._collect("_shape", *obsid)

    def get_flat_shape(self, *obsid: str | int):
        """Return the shape of the flattened values.

        Args:
            *obsid: name or index of selected observables (Default all)
        """
        vals = (
            reduce(lambda x, y: x * y, shp, 1)
            for shp in self._collect("_shape", *obsid)
        )
        return (sum(vals),)

    def get_values(self, *obsid: str | int, err: float | None = None) -> tuple:
        """Return the values of observables.

        Args:
            *obsid: name or index of selected observables (Default all)
            err:    Default observable value to be used when the evaluation failed. By
              default, an Exception is raised.

        Raises:
            Exception: Any exception raised during evaluation, unless *err* has been
              set.
        """
        return self._collect("value", *obsid, err=err)

    def get_flat_values(
        self, *obsid: str | int, err: float | None = None, order: str = "F"
    ) -> npt.NDArray[float]:
        """Return a 1-D array of Observable values.

        Args:
            *obsid: name or index of selected observables (Default all)
            err:    Default observable value to be used when the evaluation failed. By
              default, an Exception is raised.
            order:  Ordering for reshaping. See :py:func:`~numpy.reshape`
        """
        return _flatten(self._collect("value", *obsid, err=err), order=order)

    def get_weighted_values(self, *obsid: str | int, err: float | None = None) -> tuple:
        """Return the weighted values of observables.

        Args:
            *obsid: name or index of selected observables (Default all)
            err:    Default observable value to be used when the evaluation failed. By
              default, an Exception is raised.
        """
        return self._collect("weighted_value", *obsid, err=err)

    def get_flat_weighted_values(
        self, *obsid: str | int, err: float | None = None, order: str = "F"
    ) -> npt.NDArray[float]:
        """Return a 1-D array of Observable weighted values.

        Args:
            *obsid: name or index of selected observables (Default all)
            err:    Default observable value to be used when the evaluation failed. By
              default, an Exception is raised.
            order:  Ordering for reshaping. See :py:func:`~numpy.reshape`
        """
        return _flatten(self._collect("weighted_value", *obsid, err=err), order=order)

    def get_deviations(self, *obsid: str | int, err: float | None = None) -> tuple:
        """Return the deviations from target values.

        Args:
            *obsid: name or index of selected observables (Default all)
            err:    Default observable value to be used when the evaluation failed. By
              default, an Exception is raised.
        """
        return self._collect("deviation", *obsid, err=err)

    def get_flat_deviations(
        self, *obsid: str | int, err: float | None = None, order: str = "F"
    ) -> npt.NDArray[float]:
        """Return a 1-D array of deviations from target values.

        Args:
            *obsid: name or index of selected observables (Default all)
            err:    Default observable value to be used when the evaluation failed. By
              default, an Exception is raised.
            order:  Ordering for reshaping. See :py:func:`~numpy.reshape`
        """
        return _flatten(self._collect("deviation", *obsid, err=err), order=order)

    def get_weighted_deviations(
        self, *obsid: str | int, err: float | None = None
    ) -> tuple:
        """Return the weighted deviations from target values.

        Args:
            *obsid: name or index of selected observables (Default all)
            err:    Default observable value to be used when the evaluation failed. By
              default, an Exception is raised.
        """
        return self._collect("weighted_deviation", *obsid, err=err)

    def get_flat_weighted_deviations(
        self, *obsid: str | int, err: float | None = None, order: str = "F"
    ) -> npt.NDArray[float]:
        """Return a 1-D array of weighted deviations from target values.

        Args:
            *obsid: name or index of selected observables (Default all)
            err:    Default observable value to be used when the evaluation failed. By
              default, an Exception is raised.
            order:  Ordering for reshaping. See :py:func:`~numpy.reshape`
        """
        return _flatten(
            self._collect("weighted_deviation", *obsid, err=err), order=order
        )

    def get_weights(self, *obsid: str | int) -> tuple:
        """Return the weights of observables.

        Args:
            *obsid: name or index of selected observables (Default all)
        """
        return self._collect("weight", *obsid)

    def get_flat_weights(
        self, *obsid: str | int, order: str = "F"
    ) -> npt.NDArray[float]:
        """Return a 1-D array of Observable weights.

        Args:
            *obsid: name or index of selected observables (Default all)
            order:  Ordering for reshaping. See :py:func:`~numpy.reshape`
        """
        return _flatten(self._collect("weight", *obsid), order=order)

    def get_residuals(self, *obsid: str | int, err: float | None = None) -> tuple:
        """Return the residuals of observables.

        Args:
             *obsid: name or index of selected observables (Default all)
            err:    Default observable value to be used when the evaluation failed. By
              default, an Exception is raised.
        """
        return self._collect("residual", *obsid, err=err)

    def get_sum_residuals(self, *obsid: str | int, err: float | None = None) -> float:
        """Return the sum of residual values.

        Args:
            *obsid: name or index of selected observables (Default all)
            err:    Default observable value to be used when the evaluation failed. By
              default, an Exception is raised.
        """
        return sum(np.sum(res) for res in self._collect("residual", *obsid, err=err))

    shapes = property(get_shapes, doc="Shapes of all values")
    flat_shape = property(get_flat_shape, doc="Shape of the flattened values")
    values = property(get_values, doc="values of all observables")
    flat_values = property(get_flat_values, doc="1-D array of Observable values")
    weighted_values = property(
        get_weighted_values, doc="Weighted values of all observables"
    )
    flat_weighted_values = property(
        get_flat_weighted_values, doc="1-D array of Observable weigthed values"
    )
    deviations = property(get_deviations, doc="Deviations from target values")
    flat_deviations = property(
        get_flat_deviations, doc="1-D array of deviations from target value"
    )
    weighted_deviations = property(
        get_weighted_deviations, doc="Weighted deviations from target values"
    )
    flat_weighted_deviations = property(
        get_flat_weighted_deviations,
        doc="1-D array of weighted deviations from target values",
    )
    flat_weights = property(get_flat_weights, doc="1-D array of Observable weights")
    residuals = property(get_residuals, doc="Residuals of all observable")
    sum_residuals = property(get_sum_residuals, doc="Sum of all residual values")



"""Stub file for the 'atpass' extension."""

from __future__ import annotations

import numpy as np

from at.lattice import Element, Particle

_defref = np.array([], dtype=np.uint32)

def atpass(
    line: list[Element],
    r_in: np.ndarray,
    nturns: int,
    refpts: np.ndarray = _defref,
    turn: int | None = None,
    energy: float | None = None,
    particle: Particle | None = None,
    keep_counter: bool = False,
    reuse: bool = False,
    omp_num_thread: int = 0,
    losses: bool = False,
    bunch_spos=None,
    bunch_current=None,
): ...
def elempass(
    element: Element,
    r_in: np.ndarray,
    energy: float | None = None,
    particle: Particle | None = None,
): ...
def diffusion_matrix(
    element: Element,
    r_in: np.ndarray,
    energy: float | None = None,
    particle: Particle | None = None,
): ...
def reset_rng(*, rank: int = 0, seed: int | None = None) -> None: ...
def common_rng() -> float: ...
def thread_rng() -> float: ...



"""Test the legacy matching."""

from __future__ import annotations

import numpy as np
from numpy.testing import assert_allclose as assert_close
import pytest

from at.lattice import Lattice
from at.physics import linopt2
from at.deprecated import LinoptConstraints, EnvelopeConstraints
from at.deprecated import ElementVariable, match


@pytest.fixture()
def mring(hmba_lattice: Lattice):
    """Return hmba lattice with focusing sextupoles split."""
    ring = hmba_lattice.deepcopy()
    sf = ring.get_uint32_index("SF*")
    sf1 = ring[sf[0]].divide([0.5, 0.5])
    sf2 = ring[sf[1]].divide([0.5, 0.5])
    ring.pop(sf[1])
    ring.insert(sf[1], sf2[1])
    ring.insert(sf[1], sf2[0])
    ring.pop(sf[0])
    ring.insert(sf[0], sf1[1])
    ring.insert(sf[0], sf1[0])
    ring.periodicity = 1
    return ring


def test_linopt_matching(mring: Lattice):
    """Test linear optics matching."""
    # Define the location of constraint
    sf = mring.get_uint32_index("SF*")[1::2]
    center = mring.get_uint32_index("CellCenter")

    # Define the variables
    names = ["QF1*", "QD2*", "QD3*", "QF4*", "QD5*", "QF6*", "QF8*"]
    bounds = [[0, 5], [-5, 0], [-5, 0], [0, 5], [-5, 0], [0, 5], [0, 5]]
    variables = [
        ElementVariable(mring.get_uint32_index(nm), "PolynomB", index=1, name=nm)
        for nm, bnd in zip(names, bounds)
    ]

    # Define an evaluation function for the phase advance (both planes)
    # noinspection PyUnusedLocal
    def mu_diff(lindata, tune, chrom):
        delta_mu = (lindata[1].mu - lindata[0].mu) / (2 * np.pi)
        return delta_mu % 1.0

    # Define the constraints
    cst1 = LinoptConstraints(mring, method=linopt2)
    cst1.add("tunes", [0.38, 0.85], name="tunes")
    # Start of the ring
    cst1.add("alpha", [0.0, 0.0], refpts=0, name="alpha_0")
    cst1.add("dispersion", 0.0, refpts=0, index=1, name="eta'_0")
    # Center of the cell
    cst1.add("beta", 4.69999257, refpts=center, index=1, name="betay_c")
    cst1.add("alpha", [0.0, 0.0], refpts=center, name="alpha_c")
    cst1.add("dispersion", 0.0, refpts=center, index=1, name="eta'_c")
    # Focusing sextupoles
    cst1.add("beta", [5.40000677, 5.39998124], refpts=sf, index=1, name="betay_sf")
    cst1.add("alpha", [0.68000392, -0.67999686], refpts=sf, index=1, name="alphay_sf")
    cst1.add("dispersion", [0.08820002, 0.08819999], refpts=sf, index=0, name="eta_sf")
    # Phase advance
    cst1.add(mu_diff, [0.49721338, 0.48228011], refpts=sf, name="delta phi")

    # Perform the fit
    newring = match(mring, variables, (cst1,), method="lm", verbose=1)

    # check the residuals
    assert_close(cst1.evaluate(newring), 0, rtol=0.0, atol=1e-4)


def test_envelope_matching(mring: Lattice):
    """Test linear and envelope."""
    # Define the variables
    ring = mring.radiation_on(copy=True)
    names = ["QF1*", "QD2*"]
    variables = [
        ElementVariable(ring.get_uint32_index(nm), "PolynomB", index=1, name=nm)
        for nm in names
    ]
    variables += [ElementVariable([0], "Voltage", name="voltage")]
    # [0] instead of 0 because of bug in utils.py

    # Define the constraints
    lopcst = EnvelopeConstraints(ring)
    lopcst.add("tunes", [0.38, 0.85, 1.1e-4])

    # Perform the fit
    newring = match(ring, variables, (lopcst,), verbose=1)

    # check the residuals
    residual = lopcst.evaluate(newring.radiation_on(copy=True))
    assert_close(residual, 0, rtol=0.0, atol=6e-9)

    # Define the constraints
    lincst = LinoptConstraints(ring)
    lincst.add("tunes", [0.38, 0.85], name="tunes")
    lopcst = EnvelopeConstraints(ring)
    lopcst.add("tunes", 1.1e-4, index=2, name="sync. tune")

    # Perform the fit
    newring = match(ring, variables, (lincst, lopcst), verbose=1)

    # check the residuals
    linresidual = lincst.evaluate(newring)
    lopresidual = lopcst.evaluate(newring.radiation_on(copy=True))
    assert_close(linresidual, 0, rtol=0.0, atol=6e-9)
    assert_close(lopresidual, 0, rtol=0.0, atol=3e-8)



from __future__ import annotations

__all__ = [
    "lattice_track",
    "element_track",
    "internal_lpass",
    "internal_epass",
    "internal_plpass",
    "gpu_info",
    "gpu_core_count",
    "MPMode",
]

import multiprocessing
from collections.abc import Iterable
from functools import partial
from warnings import warn
from enum import Enum

import numpy

from .atpass import atpass as _atpass, elempass as _elempass, reset_rng
from .utils import fortran_align, has_collective, format_results
from .utils import initialize_lpass, disable_varelem, variable_refs
from ..lattice import AtError, AtWarning, DConstant, random
from ..lattice import Lattice, Element, Refpts, End
from ..lattice import get_uint32_index
from ..cconfig import iscuda
from ..cconfig import isopencl


class MPMode(Enum):
    """
    Multi Processing mode
    """

    CPU = 1  #: CPU multiprocessing
    GPU = 2  #: GPU multiprocessing


# Possible values for the use_mp flag, converts to use_mp,use_gpu bool pair
_mp_table = {
    False: (False, False),
    True: (True, False),
    MPMode.CPU: (True, False),
    MPMode.GPU: (False, True),
}

if iscuda() or isopencl():
    from .gpu import gpupass as _gpupass
    from .gpu import gpuinfo as _gpuinfo

_imax = numpy.iinfo(int).max
_globring: list[Element] | None = None


def _atpass_fork(seed, rank, rin, **kwargs):
    """Single forked job"""
    reset_rng(rank=rank, seed=seed)
    result = _atpass(_globring, rin, **kwargs)
    return rin, result


def _atpass_spawn(ring, seed, rank, rin, **kwargs):
    """Single spawned job"""
    reset_rng(rank=rank, seed=seed)
    result = _atpass(ring, rin, **kwargs)
    return rin, result


def _pass(ring, r_in, pool_size, start_method, seed, **kwargs):
    ctx = multiprocessing.get_context(start_method)
    # Split input in as many slices as processes
    args = enumerate(numpy.array_split(r_in, pool_size, axis=1))
    # Generate a new starting point for C RNGs
    global _globring
    _globring = ring
    if ctx.get_start_method() == "fork":
        passfunc = partial(_atpass_fork, seed, **kwargs)
    else:
        passfunc = partial(_atpass_spawn, ring, seed, **kwargs)
    # Start the parallel jobs
    with ctx.Pool(pool_size) as pool:
        results = pool.starmap(passfunc, args)
    _globring = None
    # Gather the results
    losses = kwargs.pop("losses", False)
    return format_results(results, r_in, losses)


@fortran_align
def _element_pass(element: Element, r_in, **kwargs):
    return _elempass(element, r_in, **kwargs)


@fortran_align
def _lattice_pass(
    lattice: list[Element],
    r_in,
    nturns: int = 1,
    refpts: Refpts = End,
    no_varelem=True,
    seed: int | None = None,
    use_gpu: bool = False,
    **kwargs,
):
    kwargs["reuse"] = kwargs.pop("keep_lattice", False)
    if no_varelem:
        lattice = disable_varelem(lattice)
    else:
        if sum(variable_refs(lattice)) > 0:
            kwargs["reuse"] = False
    refs = get_uint32_index(lattice, refpts)
    if seed is not None:
        reset_rng(seed=seed)
    if use_gpu:
        if not (iscuda() or isopencl()):
            raise AtError("No GPU support enabled")
        else:
            return _gpupass(lattice, r_in, nturns, refpts=refs, **kwargs)
    else:
        return _atpass(lattice, r_in, nturns, refpts=refs, **kwargs)


@fortran_align
def _plattice_pass(
    lattice: list[Element],
    r_in,
    nturns: int = 1,
    refpts: Refpts = End,
    seed: int | None = None,
    pool_size: int = None,
    start_method: str = None,
    **kwargs,
):
    verbose = kwargs.pop("verbose", False)
    refpts = get_uint32_index(lattice, refpts)
    any_collective = has_collective(lattice)
    kwargs["reuse"] = kwargs.pop("keep_lattice", False)
    rshape = r_in.shape
    if len(rshape) >= 2 and rshape[1] > 1 and not any_collective:
        if seed is None:
            seed = random.common.integers(0, high=_imax, dtype=int)
        if pool_size is None:
            pool_size = min(
                len(r_in[0]), multiprocessing.cpu_count(), DConstant.patpass_poolsize
            )
        if start_method is None:
            start_method = DConstant.patpass_startmethod
        return _pass(
            lattice,
            r_in,
            pool_size,
            start_method,
            seed=seed,
            nturns=nturns,
            refpts=refpts,
            **kwargs,
        )
    else:
        if seed is not None:
            reset_rng(seed=seed)
        if any_collective:
            warn(
                AtWarning("Collective PassMethod found: use single process"),
                stacklevel=2,
            )
        else:
            warn(
                AtWarning("no parallel computation for a single particle"), stacklevel=2
            )
        return _atpass(lattice, r_in, nturns=nturns, refpts=refpts, **kwargs)


def lattice_track(
    lattice: Iterable[Element],
    r_in,
    nturns: int = 1,
    refpts: Refpts = End,
    in_place: bool = False,
    **kwargs,
):
    """
    :py:func:`track_function` tracks particles through each element of a
    lattice or throught a single Element calling the element-specific
    tracking function specified in the Element's *PassMethod* field.

    Usage:
      >>> lattice_track(lattice, r_in)
      >>> lattice.track(r_in)

    Parameters:
        lattice: list of elements
        r_in: (6, N) array: input coordinates of N particles.
          *r_in* is modified in-place only if *in_place* is
          :py:obj:`True` and reports the coordinates at
          the end of the element. For the best efficiency, *r_in*
          should be given as F_CONTIGUOUS numpy array.

    Keyword arguments:
        nturns: number of turns to be tracked
        refpts: Selects the location of coordinates output.
          See ":ref:`Selecting elements in a lattice <refpts>`"
        in_place (bool): If True *r_in* is modified in-place and
          reports the coordinates at the end of the element.
          (default: False)
        seed (int | None): Seed for the random generators. If None (default)
          continue the sequence
        keep_lattice (bool):    Use elements persisted from a previous
          call. If :py:obj:`True`, assume that the lattice has not changed
          since the previous call.
        keep_counter (bool):    Keep the turn number from the previous
          call.
        turn (int):             Starting turn number. Ignored if
          *keep_counter* is :py:obj:`True`. The turn number is necessary to
          compute the absolute path length used in RFCavityPass.
        losses (bool):          Boolean to activate loss maps output
        omp_num_threads (int):  Number of OpenMP threads
          (default: automatic)
        use_mp (bool | MPMode):  Flag to activate multiprocessing
          (default: False)
        pool_size:              number of processes used when
          *use_mp* is :py:obj:`True` or :py:attr:`MPMode.CPU`. If None,
          ``min(npart,nproc)`` is used. It can be globally set using the
          variable *at.lattice.DConstant.patpass_poolsize*
        gpu_pool: List of GPU to use when *use_mp* is :py:attr:`MPMode.GPU`.
          If none, first GPU is selected.
        start_method:           python multiprocessing start method.
          :py:obj:`None` uses the python default that is considered safe.
          Available values: ``'fork'``, ``'spawn'``, ``'forkserver'``.
          Default for linux is ``'fork'``, default for macOS and  Windows is
          ``'spawn'``. ``'fork'`` may be used on macOS to speed up the
          calculation or to solve Runtime Errors, however it is considered
          unsafe. Used only when *use_mp* is :py:obj:`True`. It can be globally
          set using the variable *at.lattice.DConstant.patpass_startmethod*

    The following keyword arguments overload the lattice values

    Keyword arguments:

        particle (Optional[Particle]): circulating particle.
          Default: :code:`lattice.particle` if existing,
          otherwise :code:`Particle('relativistic')`
        energy (Optiona[float]): lattice energy. Default 0.
        unfold_beam (bool): Internal beam folding activate, this
          assumes the input particles are in bucket 0, works only
          if all bucket see the same RF Voltage.
          Default: :py:obj:`True`

    If *energy* is not available, relativistic tracking if forced,
    *rest_energy* is ignored.

    Returns:
        r_out: (6, N, R, T) array containing output coordinates of N particles
          at R reference points for T turns
        trackparam: A dictionary containing tracking input parameters with the
          following keys:

          ==============    ===================================================
          **npart**         number of particles
          **rout**          final particle coordinates
          **turn**          starting turn
          **refpts**        array of index where particle coordinate are saved
                            (only for lattice tracking)
          **nturns**        number of turn
          ==============    ===================================================

        trackdata: A dictionary containing tracking data with the following
          keys:

          ==============    ===================================================
          **loss_map**:     recarray containing the loss_map (only for lattice
                            tracking)
          ==============    ===================================================


          The **loss_map** is filled only if *losses* is :py:obj:`True`,
          it contains the following keys:

          ==============    ===================================================
          **islost**        (npart,) bool array indicating lost particles
          **turn**          (npart,) int array indicating the turn at
                            which the particle is lost
          **element**       (npart,) int array indicating the element at
                            which the particle is lost
          **coord**         (npart, 6) float array giving the coordinates at
                            which the particle is lost (zero for surviving
                            particles)
          ==============    ===================================================


    .. note::

       * :pycode:`track_function(lattice, r_in, refpts=len(line))` is the same
         as :pycode:`track_function(lattice, r_in)` since the reference point
         len(line) is the exit of the last element.
       * :pycode:`track_function(lattice, r_in, refpts=0)` is a copy of *r_in*
         since the reference point 0 is the entrance of the first element.
       * To resume an interrupted tracking (for instance to get intermediate
         results), one must use one of the *turn* or *keep_counter*
         keywords to ensure the continuity of the turn number.
       * For multiparticle tracking with large number of turn the size of
         *r_out* may increase excessively. To avoid memory issues
         :pycode:`track_function(lattice, r_in, refpts=None, in_place=True)`
         can be used. An empty list is returned and the tracking results of
         the last turn are stored in *r_in*.
       * To model buckets with different RF voltage :pycode:`unfold_beam=False`
         has to be used. The beam can be unfolded using the function
         :py:func:`.unfold_beam`. This function takes into account
         the true voltage in each bucket and distributes the particles in the
         bunches defined by :code:`ring.fillpattern` using a 6D orbit search.
    """

    use_mp = kwargs.pop("use_mp", False)
    use_mp, use_gpu = _mp_table[use_mp]

    trackdata = {}
    trackparam = {}
    part_kw = ["energy", "particle"]
    try:
        npart = numpy.shape(r_in)[1]
    except IndexError:
        npart = 1

    [trackparam.update((kw, kwargs.get(kw))) for kw in kwargs if kw in part_kw]
    trackparam.update({"npart": npart})

    if not in_place:
        r_in = r_in.copy()

    lattice = initialize_lpass(lattice, nturns, kwargs)
    ldtype = [
        ("islost", numpy.bool_),
        ("turn", numpy.uint32),
        ("elem", numpy.uint32),
        ("coord", numpy.float64, (6,)),
    ]
    loss_map = numpy.recarray((npart,), ldtype)
    lat_kw = ["turn"]
    [trackparam.update((kw, kwargs.get(kw))) for kw in kwargs if kw in lat_kw]
    trackparam.update({"refpts": get_uint32_index(lattice, refpts), "nturns": nturns})

    start_method = kwargs.pop("start_method", None)
    pool_size = kwargs.pop("pool_size", None)
    if use_mp:
        kwargs.update({"pool_size": pool_size, "start_method": start_method})
        rout = _plattice_pass(lattice, r_in, nturns=nturns, refpts=refpts, **kwargs)
    else:
        rout = _lattice_pass(
            lattice,
            r_in,
            nturns=nturns,
            refpts=refpts,
            use_gpu=use_gpu,
            no_varelem=False,
            **kwargs,
        )

    if kwargs.get("losses", False):
        rout, lm = rout
        lm["coord"] = lm["coord"].T
        for k, v in lm.items():
            loss_map[k] = v

    trackdata.update({"loss_map": loss_map})
    trackparam.update({"rout": r_in})

    return rout, trackparam, trackdata


def element_track(element: Element, r_in, in_place: bool = False, **kwargs):
    """
    :py:func:`element_track` tracks particles through one element of a
    calling the element-specific tracking function specified in the
    Element's *PassMethod* field

    Usage:
      >>> element_track(element, r_in)
      >>> element.track(r_in)

    Parameters:
        element: element to track through
        r_in: (6, N) array: input coordinates of N particles.
          For the best efficiency, *r_in*
          should be given as F_CONTIGUOUS numpy array.

    Keyword arguments:
        in_place (bool): If True *r_in* is modified in-place and
          reports the coordinates at the end of the element.
          (default: False)
        omp_num_threads (int):  Number of OpenMP threads
          (default: automatic)
        particle (Optional[Particle]): circulating particle.
          Default: :code:`lattice.particle` if existing,
          otherwise :code:`Particle('relativistic')`
        energy (Optiona[float]): lattice energy. Default 0.

    Returns:
        r_out: (6, N, R, T) array containing output coordinates of N particles
          at R reference points for T turns
    """
    if not in_place:
        r_in = r_in.copy()

    rout = _element_pass(element, r_in, **kwargs)
    return rout


def gpu_core_count(gpuPool: list[int] | None):
    """
    :py:func:`gpu_core_count` returns number of GPU core.

    Parameters:
        gpuPool: List of selected GPUs. If None, first GPU is selected.
    """
    gpus = gpu_info()
    nbgpu = len(gpus)
    nbcore = 0
    if gpuPool is None:
        gpuPool = [0]
    for g in gpuPool:
        if g < 0 or g >= nbgpu:
            raise AtError(
                "Invalid GPU id " + str(g) + ", must be in [0.." + str(nbgpu - 1) + "]"
            )
        nbcore += gpus[g][2]  # Compute unit number
    return nbcore


def gpu_info():
    """
    :py:func:`gpu_info` returns list of GPU present on the system and their
    corresponding information. If GPU support is not enabled or if no capable device
    are present on the system, an empty list is returned.

    Returns:
        gpu: [gpu name,hardware version (CUDA device),compute unit number,platform].
          The number of compute units is not the so-called number of "CUDA cores".
          The number of threads that can be executed simultaneously depends on the
          hardware and on the type of used instructions. For best performance, it is
          recommended to track a number of particles which is multiple of 64 (or 128)
          times the number of compute units.
    """
    if iscuda() or isopencl():
        return _gpuinfo()
    else:
        return []


internal_lpass = _lattice_pass
internal_epass = _element_pass
internal_plpass = _plattice_pass
Lattice.track = lattice_track
Element.track = element_track
